r"""
Network abstraction

Doctest Examples::

  # Client: Twisted syntax
  >>> ClientEndpoint("tcp:host=example.com:port=1234", protocol="http").geturl()
  'http://example.com:1234'
  >>> ClientEndpoint("tls:host=example.com:port=1234", protocol="http").geturl()
  'https://example.com:1234'

  # Client: URL syntax
  >>> ce = ClientEndpoint("http://example.com:1234/")
  >>> ce.description
  'tcp:host=example.com:port=1234'
  >>> ClientEndpoint("https://example.com:1234").description
  'tls:host=example.com:port=1234'
  >>> ClientEndpoint("https://example.com:1234").geturl(with_user=True, path="api", query={"a": "b", "c": "d"})
  'https://example.com:1234/api?a=b&c=d'
  >>> ClientEndpoint("https://foo@example.com:1234").geturl(with_user=False, path="api", query={"a": "b", "c": "d"})
  'https://example.com:1234/api?a=b&c=d'
  >>> ClientEndpoint("https://bar@example.com:1234").geturl(with_user=True, path="api", query={"a": "b", "c": "d"})
  'https://bar@example.com:1234/api?a=b&c=d'
"""

import http.client
import logging
import re
import ssl
import urllib.error
import urllib.parse
import urllib.request

import twisted.internet.endpoints
from twisted.internet import reactor

from mini_buildd import config, util

LOG = logging.getLogger(__name__)


class TwistedEndpoint():
    @classmethod
    def _escape(cls, string):
        return string.replace(":", r"\:")

    @classmethod
    def _unescape(cls, string):
        return string.replace(r"\:", ":")

    @classmethod
    def _opt(cls, key, value):
        return f":{key}={cls._escape(value)}" if value else ""

    SUPPORTED_TYPES = ["ssl", "tls", "tcp6", "tcp", "unix"]

    def __init__(self, description):
        self.description = description

        params = [self._unescape(p) for p in re.split(r"(?<!\\):", description)]
        self._type = params[0]
        if self._type not in self.SUPPORTED_TYPES:
            raise util.HTTPBadRequest(f"Unsupported endpoint type: {self._type} (twisted types supported: {','.join(self.SUPPORTED_TYPES)})")

        self.options = {}
        for p in params:
            key = p.partition("=")
            if key[1]:
                self.options[key[0]] = key[2]
        # LOG.debug(f"Twisted endpoint parsed: {self.description}")

    def is_ssl(self):
        return self._type in ["ssl", "tls"]


class Endpoint(TwistedEndpoint):
    def netloc(self, with_user=True, user=None):
        user = self._user if user is None else user
        host = self.options.get("host", self.options.get("interface", config.HOSTNAME_FQDN))

        port = self.options.get("port")
        return (f"{user}@" if (user is not None and with_user) else "") + host + (f":{port}" if port else "")

    def __init__(self, description, protocol):
        super().__init__(description)
        self._user = None

        self._url = urllib.parse.ParseResult(scheme=protocol + ("s" if self.is_ssl() else ""),
                                             netloc=self.netloc(),
                                             path="", params="", query="", fragment="")
        # LOG.debug(f"Endpoint initialized: {self.description}")

    def __str__(self):
        return self.geturl()

    def get_user(self):
        return self._user

    def set_user(self, user):
        self._user = user

    def hopo(self):
        return f"{self._url.hostname}:{self.options.get('port')}"

    def scheme(self):
        return self._url.scheme

    def geturl(self, with_user=False, user=None, path="", query=None, relative=False):
        return urllib.parse.ParseResult(scheme="" if relative else self.scheme(),
                                        netloc="" if relative else self.netloc(with_user=with_user, user=user),
                                        path=path,
                                        params="",
                                        query=urllib.parse.urlencode({} if query is None else query, doseq=False),
                                        fragment="").geturl()


class ServerEndpoint(Endpoint):
    """
    Twisted-style Network Server Endpoint

    See: https://twistedmatrix.com/documents/current/core/howto/endpoints.html#servers

    Notation Examples::

      'tcp6:port=8066'                                     : Unencrypted, bind on all interfaces.
      'ssl:port=8066:privateKey=KEY_PATH:certKey=CERT_PATH': Encrypted, bind on all interfaces.
      'tcp6:interface=localhost:port=8066'                 : Unencrypted, bind on localhost only.

    """

    def __init__(self, description, protocol):
        twisted.internet.endpoints.serverFromString(reactor, description)  # Syntax check only for now
        super().__init__(description, protocol)

    def get_certificate(self):
        cert_key_file = self.options.get("certKey")
        if cert_key_file:
            with util.fopen(cert_key_file) as f:
                return f.read()
        return ""


class ClientEndpoint(Endpoint):
    """
    Twisted-style Network Client Endpoint

    See: https://twistedmatrix.com/documents/current/core/howto/endpoints.html#clients

    Notation Examples::

      'tcp:host=localhost:port=8066': Connect to localhost, unencrypted.
      'tls:host=localhost:port=8066': Connect to localhost, encrypted.
      'http://localhost:8066'       : Extra URL-style syntax (prefer twisted-style in saved configs, though).
      'localhost:8066'              : Compatibility host:port syntax (don't use -- might eventually go away).
    """

    def __init__(self, description, protocol="http"):
        user = None
        if re.match(r".*://.*", description):
            parsed_url = urllib.parse.urlsplit(description)
            if parsed_url.password:
                raise util.HTTPBadRequest("We don't allow to give password in URL")

            typ, default_port, _protocol = {"http": ("tcp", "80", "http"),
                                            "https": ("tls", "443", "http"),
                                            "ftp": ("tcp", "21", "ftp"),
                                            "ftps": ("tls", "990", "ftp")}[parsed_url.scheme]
            _user_host, dummy, port = parsed_url.netloc.rpartition(":")
            host = _user_host.rpartition("@")[2]

            protocol = _protocol if protocol is None else protocol
            description = typ + self._opt("host", host) + self._opt("port", default_port if port is None else port)
            user = parsed_url.username

        twisted.internet.endpoints.clientFromString(reactor, description)  # Syntax check only for now
        super().__init__(description, protocol)
        self.set_user(user)

        # Login/out handling
        self.set_user(user)
        self._anon_opener = urllib.request.build_opener()
        self._opener = self._anon_opener

    def __str__(self):
        return f"{self.geturl(with_user=True)}({'logged in' if self.logged_in() else 'logged out'})"

    def get_certificate(self):
        return ssl.get_server_certificate((self.options.get("host"), self.options.get("port")))

    def login(self, password):
        login_url = self.geturl(path=config.URIS["accounts"]["login"].join())
        next_url = self.geturl(path=config.URIS["accounts"]["null"].join())

        # Create cookie-enabled opener
        cookie_handler = urllib.request.HTTPCookieProcessor()
        opener = urllib.request.build_opener(cookie_handler)

        # Retrieve login page
        opener.open(login_url)
        opener.addheaders = [("Referer", self.geturl())]

        # Find "csrftoken" in cookiejar
        csrf_cookies = [c for c in cookie_handler.cookiejar if c.name == "csrftoken"]
        if len(csrf_cookies) != 1:
            raise util.HTTPBadRequest(f"{len(csrf_cookies)} csrftoken cookies found in login pages (need exactly 1)")

        # Login via POST request
        response = opener.open(
            login_url,
            bytes(urllib.parse.urlencode({"username": self._user,
                                          "password": password,
                                          "csrfmiddlewaretoken": csrf_cookies[0].value,
                                          "this_is_the_login_form": "1",
                                          "next": next_url}),
                  encoding=config.CHAR_ENCODING))

        # If successful, next url of the response must match
        if response.geturl() != next_url:
            raise util.HTTPUnauthorized(f"Login for user '{self._user}' failed")

        # Logged in: Install opener
        self._opener = opener

    def logout(self):
        self._opener = self._anon_opener

    def logged_in(self):
        return self._opener != self._anon_opener

    def urlopen(self, url, timeout=None):
        return self._opener.open(url, timeout=timeout)

    def http_connect(self):
        return (http.client.HTTPSConnection if self.is_ssl() else http.client.HTTPConnection)(self.options.get("host"), port=self.options.get("port"))


def detect_apt_cacher_ng(url="http://localhost:3142"):
    """Little heuristic helper for the "local archives" wizard"""
    try:
        with urllib.request.urlopen(url) as _:
            pass
    except urllib.error.HTTPError as e:
        if e.code == 406 and re.findall(r"apt.cacher.ng", e.file.read().decode(config.CHAR_ENCODING), re.IGNORECASE):
            return url
    except BaseException:
        pass
    return None



import fnmatch
import logging
import os
import stat

import pyftpdlib.authorizers
import pyftpdlib.handlers
import pyftpdlib.servers

from mini_buildd import changes, config, threads, util

LOG = logging.getLogger(__name__)


class FtpDHandlerMixin():
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._mbd_files_received = []

    def on_file_received(self, file):
        """Make any incoming file read-only as soon as it arrives; avoids overriding uploads of the same file"""
        os.chmod(file, stat.S_IRUSR | stat.S_IRGRP)
        self._mbd_files_received.append(file)
        LOG.debug("File received: %s", file)

    def on_incomplete_file_received(self, file):
        LOG.warning("Incomplete file received: %s", file)
        self._mbd_files_received.append(file)

    def on_disconnect(self):
        for file_path in (f for f in self._mbd_files_received if fnmatch.fnmatch(f, "*.changes")):
            try:
                changes.incoming2queue(file_path)
            except Exception as e:
                util.log_exception(LOG, f"Invalid incoming changes: {file_path}", e)

        # All received files should be queued by now, any left-over in received files is cruft
        for file_path in (f for f in self._mbd_files_received if os.path.exists(f)):
            LOG.warning("Removing cruft incoming: %s", file_path)
            os.remove(file_path)


class FtpDHandler(FtpDHandlerMixin, pyftpdlib.handlers.FTPHandler):
    pass


class FtpsDHandler(FtpDHandlerMixin, pyftpdlib.handlers.TLS_FTPHandler):
    pass


class FtpD(threads.Thread):
    HANDLER_OPTIONS_NAMES = ["passive_ports"]
    HANDLER_OPTIONS_USAGE = f"""\
ftpd handler options (see ``https://pyftpdlib.readthedocs.io/en/latest/api.html#control-connection``).

Format is ``name0=value0;name1=value1...``.

Supported options: ``{','.join(HANDLER_OPTIONS_NAMES)}``

Divergent string values:

``passive_ports=<min_port>-<max_port>``
"""

    @classmethod
    def parse_handler_options(cls, str_options):
        try:
            parsed = {o.split("=")[0]: o.split("=")[1] for o in (str_options.split(";") if str_options else [])}
        except BaseException as e:
            raise util.HTTPBadRequest(f"ftpd: Error parsing handler options syntax: {str_options}") from e

        try:
            result = {}
            for name, str_value in parsed.items():
                if name not in cls.HANDLER_OPTIONS_NAMES:
                    raise util.HTTPBadRequest(f"ftpd: Unhandled handler option: {name}")
                # Transfer values where needed
                result[name] = {
                    "passive_ports": lambda v: list(range(int(v.split("-")[0]), int(v.split("-")[1]) + 1)),
                }.get(name, lambda v: v)(str_value)
            return result
        except BaseException as e:
            raise util.HTTPBadRequest(f"ftpd: Error parsing handler options values: {e}") from e

    def __init__(self, endpoint):
        super().__init__()
        self.endpoint = endpoint
        self.ftpd = None

    def __str__(self):
        return f"{super().__str__()} ({self.endpoint.geturl()})"

    def bind(self, handler_options):
        if self.endpoint.is_ssl():
            handler = FtpsDHandler
            handler.certfile = self.endpoint.options.get("certKey")
            handler.keyfile = self.endpoint.options.get("privateKey")
            handler.tls_control_required = True
            handler.tls_data_required = True
        else:
            handler = FtpDHandler

        handler.authorizer = pyftpdlib.authorizers.DummyAuthorizer()
        handler.authorizer.add_anonymous(homedir=config.ROUTES["home"].path.join(), perm="")
        handler.authorizer.override_perm(username="anonymous", directory=config.ROUTES["incoming"].path.join(), perm="elrw")

        options = self.parse_handler_options(handler_options)
        for name, value in options.items():
            LOG.info("ftpd: Setting handler option: %s=%s", name, value)
            setattr(handler, name, value)

        handler.banner = f"mini-buildd {util.__version__} ftp server ready (pyftpdlib {pyftpdlib.__ver__})."

        try:
            self.ftpd = pyftpdlib.servers.FTPServer((self.endpoint.options.get("interface"), self.endpoint.options.get("port")), handler)
        except OSError as e:
            raise util.HTTPUnavailable(f"OS error: {e}")

    def mbd_run(self):
        self.ftpd.serve_forever(timeout=1.0)

    def shutdown(self):
        if self.ftpd is not None:
            self.ftpd.close_all()



# Generated by Django 4.2.9 on 2024-01-11 16:27

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('mini_buildd', '0023_alter_distribution_chroot_setup_script_and_more'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='daemon',
            name='keep_build_results',
        ),
        migrations.RemoveField(
            model_name='daemon',
            name='wait_for_build_results',
        ),
    ]



import logging
import os
import queue
from contextlib import closing

from mini_buildd import changes, config, daemon, events, net, sbuild, threads, util

LOG = logging.getLogger(__name__)


class Uploader(threads.EventThread):
    def __init__(self):
        super().__init__()

        self.failed = {}
        for bres_entry in (f for f in config.ROUTES["builds"].path.ifiles("", ".changes") if changes.Buildresult.match(f.path)):
            LOG.debug("Checking buildresult: %s", bres_entry.path)
            try:
                bres = changes.Buildresult(bres_entry.path)
                if not bres.uploaded():
                    LOG.warning("Found upload-pending buildresult: %s", bres_entry.path)
                    self.queue.put(bres)
            except Exception as e:
                LOG.error("Bogus buildresult (ignoring): %s: %s", bres_entry.path, e)

    def run_event(self, event):
        with self.lock:
            try:
                event.upload(net.ClientEndpoint(event.cget("Upload-To"), protocol="ftp"))
                self.failed.pop(event.bkey, None)
            except BaseException as e:
                util.log_exception(LOG, f"Upload to {event.cget('Upload-To')} failed (will retry)", e)
                self.failed[event.bkey] = event

    def retry_failed(self):
        with self.lock:
            for bres in self.failed.values():
                self.queue.put(bres)


class Build(threads.DeferredThread):
    def __init__(self, limiter, breq, uploader_queue):
        # Note: If the constructor fails, no buildresult would be uploaded (and packaging would hang). Keep it simple && be sure this does not fail on 'normal' error conditions
        self.breq = breq
        self.uploader_queue = uploader_queue
        self.sbuild = sbuild.SBuild(self.breq, ["--keyid", daemon.get().gnupg.get_first_sec_key()])
        super().__init__(limiter=limiter)

    def __str__(self):
        return f"{super().__str__()} ({self.breq.bkey})"

    def cancel(self, who):
        if self.is_alive():
            self.sbuild.cancel(who)

    def run_deferred(self):
        if self._shutdown is config.SHUTDOWN:
            raise util.HTTPShutdown()

        daemon.get().events.log(events.Type.BUILDING, self.breq)
        try:
            # Verify buildrequest
            with closing(daemon.RemotesKeyring()) as gpg:
                gpg.verify(self.breq.file_path)

            # Run sbuild
            self.sbuild.run()

            build_changes_file = os.path.join(self.breq.dir_path(), self.breq.dfn.changes(arch=self.breq["architecture"]))
            self.sbuild.bres.save_to(self.breq.dir_path(), changes.Base(build_changes_file).tar() if os.path.exists(build_changes_file) else None)
        except BaseException as e:
            util.log_exception(LOG, f"Internal-Error: {self}", e)
            self.sbuild.bres.cset("Internal-Error", str(util.e2http(e)))
            self.sbuild.bres.save_to(self.breq.dir_path())

        daemon.get().events.log(events.Type.BUILT, self.sbuild.bres)
        self.uploader_queue.put(self.sbuild.bres)


class Builder(threads.EventThread):
    def __init__(self, max_parallel_builds):
        self.limiter = queue.Queue(maxsize=max_parallel_builds)
        self.builds = {}
        self.uploader = Uploader()
        super().__init__(subthreads=[self.uploader])

    def __str__(self):
        return f"{super().__str__()} ({self.limiter.maxsize} parallel builds)"

    def queued(self):
        return sum(map(lambda b: 1 if b.is_alive() else 0, self.builds.values()))

    def running(self):
        return sum(map(lambda b: 1 if (b.is_alive() and b.running is not None) else 0, self.builds.values()))

    def load(self):
        return round(float(self.queued() / self.limiter.maxsize), 2)

    def clear(self):
        with self.lock:
            for key in [key for key, build in self.builds.items() if not build.is_alive()]:  # Iterate on copy, del on actual dict
                self.builds[key].join()
                del self.builds[key]

    def join(self, timeout=None):
        for build in self.builds.values():
            build.cancel("SHUTDOWN")
            build.join()
        super().join()

    def run_event(self, event):
        self.clear()
        build = Build(self.limiter, event, self.uploader.queue)
        self.builds[build.breq.bkey] = build
        build.start()

    def cancel(self, bkey, who):
        self.clear()
        build = self.builds.get(bkey)
        if build is None:
            raise util.HTTPBadRequest(f"No such active build: {bkey}")
        build.cancel(who)



# Generated by Django 4.2.8 on 2024-01-03 15:39

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('mini_buildd', '0001_initial'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='aptkey',
            name='pickled_data',
        ),
        migrations.RemoveField(
            model_name='architecture',
            name='pickled_data',
        ),
        migrations.RemoveField(
            model_name='architectureoption',
            name='pickled_data',
        ),
        migrations.RemoveField(
            model_name='archive',
            name='pickled_data',
        ),
        migrations.RemoveField(
            model_name='chroot',
            name='pickled_data',
        ),
        migrations.RemoveField(
            model_name='component',
            name='pickled_data',
        ),
        migrations.RemoveField(
            model_name='daemon',
            name='pickled_data',
        ),
        migrations.RemoveField(
            model_name='distribution',
            name='pickled_data',
        ),
        migrations.RemoveField(
            model_name='emailaddress',
            name='pickled_data',
        ),
        migrations.RemoveField(
            model_name='layout',
            name='pickled_data',
        ),
        migrations.RemoveField(
            model_name='prioritysource',
            name='pickled_data',
        ),
        migrations.RemoveField(
            model_name='remote',
            name='pickled_data',
        ),
        migrations.RemoveField(
            model_name='repository',
            name='pickled_data',
        ),
        migrations.RemoveField(
            model_name='source',
            name='pickled_data',
        ),
        migrations.RemoveField(
            model_name='subscription',
            name='pickled_data',
        ),
        migrations.RemoveField(
            model_name='suite',
            name='pickled_data',
        ),
        migrations.RemoveField(
            model_name='suiteoption',
            name='pickled_data',
        ),
        migrations.RemoveField(
            model_name='uploader',
            name='pickled_data',
        ),
    ]



# Generated by Django 4.2.8 on 2024-01-03 16:15

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('mini_buildd', '0004_remove_distribution_apt_allow_unauthenticated'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='daemon',
            name='smtp_server',
        ),
    ]



"""
Distribution setups && support

Dist-like variable naming convention
====================================
* ``diststr``: Instance of ``str``: "buster-test-unstable"
* ``dist``: Instance of ``mini_buildd.dist.Dist``: diststr parsed && support
* ``distribution``: Instance of ``mini_buildd.model.distribution.Distribution``: Configured distribution
"""

import abc
import enum
import functools
import logging
import re

import debian.debian_support
import distro_info

from mini_buildd import call, util

LOG = logging.getLogger(__name__)


class Archs:
    """Deliver lists of Debian architectures the current system can handle"""

    _NATIVE_ARCHS_MAP = {"amd64": ["i386"]}

    @classmethod
    def _atwarn(cls, e):
        LOG.warning("Please install ``arch-test`` to get a complete list of usable archs: %s", e)

    @classmethod
    def native(cls):
        try:
            return call.Call(["arch-test", "-n"]).check().stdout.split()
        except Exception as e:
            cls._atwarn(e)
            native = call.Call(["dpkg", "--print-architecture"]).check().stdout.strip()
            return [native] + cls._NATIVE_ARCHS_MAP.get(native, [])

    @classmethod
    def available(cls):
        try:
            return call.Call(["arch-test"]).check().stdout.split()
        except Exception as e:
            cls._atwarn(e)
            return cls.native()


class DistroInfo(distro_info.DistroInfo):
    def mbd_origin(self):
        return self._distro

    def mbd_release(self, codename):
        for r in self._releases:
            if r.series == codename:
                return r
        return None

    @abc.abstractmethod
    def mbd_lts(self):
        pass


class DebianDistroInfo(DistroInfo, distro_info.DebianDistroInfo):
    def mbd_lts(self):
        """
        .. tip:: What does ``LTS`` (mini-buildd speak) include for Debian?

           For mini-buildd, ``LTS`` means ``Debian Long Term Support (LTS)`` including ``Debian Extended Long Term Support (ELTS)``.

           See: https://www.debian.org/lts/, https://wiki.debian.org/LTS/Extended
        """
        return self.lts_supported() + self.elts_supported()


class UbuntuDistroInfo(DistroInfo, distro_info.UbuntuDistroInfo):
    def mbd_lts(self):
        """
        .. tip:: What does ``LTS`` (mini-buildd speak) include for Ubuntu?

           For mini-buildd, ``LTS`` means ``Ubuntu Long Term Support (LTS)`` including ``Ubuntu Extended Security Maintenance (ESM)``.

           See: https://ubuntu.com/about/release-cycle
        """
        return self.supported_esm()  # this includes LTS


#: Available distro infos
DISTRO_INFO = {
    "Debian": DebianDistroInfo(),
    "Ubuntu": UbuntuDistroInfo(),
}

APT_KEYS = {
    "Debian": {
        "trixie": {
            "archive": "225629DF75B188BD",    # Debian Archive Automatic Signing Key (13/trixie) <ftpmaster@debian.org>
            "release": "762F67A0B2C39DE4",    # Debian Stable Release Key (13/trixie) <debian-release@lists.debian.org>
            "security": "9904613D4CCE68C6",   # Debian Security Archive Automatic Signing Key (13/trixie) <ftpmaster@debian.org>
        },
        "bookworm": {
            "archive": "B7C5D7D6350947F8",    # Debian Archive Automatic Signing Key (12/bookworm) <ftpmaster@debian.org>
            "release": "F8D2585B8783D481",    # Debian Stable Release Key (12/bookworm) <debian-release@lists.debian.org>
            "security": "254CF3B5AEC0A8F0",   # Debian Security Archive Automatic Signing Key (12/bookworm) <ftpmaster@debian.org>
        },
        "bullseye": {
            "archive": "73A4F27B8DD47936",    # Debian Archive Automatic Signing Key (11/bullseye) <ftpmaster@debian.org>
            "release": "605C66F00D6C9793",    # Debian Stable Release Key (11/bullseye) <debian-release@lists.debian.org>
            "security": "A48449044AAD5C5D",   # Debian Security Archive Automatic Signing Key (11/bullseye) <ftpmaster@debian.org>
        },
        "buster": {
            "archive": "DC30D7C23CBBABEE",    # Debian Archive Automatic Signing Key (10/buster) <ftpmaster@debian.org>
            "release": "DCC9EFBF77E11517",    # Debian Stable Release Key (10/buster) <debian-release@lists.debian.org>
            "security": "4DFAB270CAA96DFA",   # Debian Security Archive Automatic Signing Key (10/buster) <ftpmaster@debian.org>
        },
        "stretch": {
            "archive": "E0B11894F66AEC98",    # Debian Archive Automatic Signing Key (9/stretch) <ftpmaster@debian.org>  (subkey 04EE7237B7D453EC)
            "release": "EF0F382A1A7B6500",    # Debian Stable Release Key (9/stretch) <debian-release@lists.debian.org>
            "security": "EDA0D2388AE22BA9",   # Debian Security Archive Automatic Signing Key (9/stretch) <ftpmaster@debian.org>
        },
        "jessie": {
            "archive": "7638D0442B90D010",    # Debian Archive Automatic Signing Key (8/jessie) <ftpmaster@debian.org>
            "release": "CBF8D6FD518E17E1",    # Jessie Stable Release Key <debian-release@lists.debian.org>
            "security": "9D6D8F6BC857C906",   # Debian Security Archive Automatic Signing Key (8/jessie) <ftpmaster@debian.org>
        },
        "wheezy": {
            "archive": "8B48AD6246925553",    # Debian Archive Automatic Signing Key (7.0/wheezy) <ftpmaster@debian.org>
            "release": "6FB2A1C265FFB764",    # Wheezy Stable Release Key <debian-release@lists.debian.org>
        },
        "squeeze": {
            "archive": "AED4B06F473041FA",    # Debian Archive Automatic Signing Key (6.0/squeeze) <ftpmaster@debian.org>
            "release": "64481591B98321F9",    # Squeeze Stable Release Key <debian-release@lists.debian.org>
        },
        "lenny": {
            "archive": "9AA38DCD55BE302B",    # Debian Archive Automatic Signing Key (5.0/lenny) <ftpmaster@debian.org>
            "release": "4D270D06F42584E6",    # Lenny Stable Release Key <debian-release@lists.debian.org>
        },
    },
    "Ubuntu": {
        "2018": "871920D1991BC93C",   # Ubuntu Archive Automatic Signing Key (2018) <ftpmaster@ubuntu.com>
        "2012": "3B4FE6ACC0B21F32",   # Ubuntu Archive Automatic Signing Key (2012) <ftpmaster@ubuntu.com>
        "2004": "40976EAF437D05B5",   # Ubuntu Archive Automatic Signing Key <ftpmaster@ubuntu.com>
    },
    "PureOS": {
        "repository": "2B4A53F2B41CE072",   # Purism PureOS Archive (PureOS Archive Signing Key) <sysadmin@puri.sm>
    },
}

SUITE_SETUP = {
    "stable": {
        "options": {
            "uploadable": False,
        },
    },
    "hotfix": {
        "migrates_to": "stable",
    },
    "testing": {
        "migrates_to": "stable",
        "options": {
            "uploadable": False,
        },
    },
    "unstable": {
        "migrates_to": "testing",
        "options": {
            "build_keyring_package": True,
        },
    },
    "snapshot": {
        "options": {
            "experimental": True,
        },
    },
    "experimental": {
        "options": {
            "experimental": True,
            "but_automatic_upgrades": False,
        },
    },
}

SETUP = {
    "defaults": {  # Plain default values for model fields
        "distribution": {
            "sbuild_config_blocks": "set-default-path ccache",
            "sbuild_setup_blocks": "apt-setup apt-update ccache eatmydata",
            "lintian_extra_options": "--info --suppress-tags bad-distribution-in-changes-file,bogus-mail-host,bogus-mail-host-in-debian-changelog",
            "lintian_warnfail_options": "--fail-on error,warning",
        },
    },
    "layout": {  # Layout Default
        "Default": {
            "suites": SUITE_SETUP,
            "rollback": {"stable": 6, "hotfix": 4, "testing": 3, "unstable": 9, "snapshot": 12, "experimental": 6},
        },
        "Default (no rollbacks)": {
            "suites": SUITE_SETUP,
            "rollback": {},
        },
        "Debian Developer": {
            "suites": {s: c for s, c in SUITE_SETUP.items() if s in ["stable", "testing", "unstable", "experimental"]},
            "rollback": {},
            "options": {
                "mandatory_version_regex": ".*",
                "experimental_mandatory_version_regex": ".*",
                "meta_distributions": {"unstable": "sid-unstable", "experimental": "sid-experimental"},
            },
        },
    },
    "origin": {  # Custom setup for actual distributons
        "Debian": {
            #: "Usually used" paths on a mirror. Can be used to guess individual archive URLS from a base mirror URL.
            "archive_paths": ["debian", "debian-security", "debian-ports", "debian-archive/debian", "debian-archive/debian-security", "debian-archive/debian-backports"],

            #: Canonical (internet) archives
            "archive": [
                "http://ftp.debian.org/debian/",                 # Debian (release, updates, proposed-updates and backports)
                "http://deb.debian.org/debian/",                 # alternate: CDN

                "http://security.debian.org/debian-security/",   # Debian Security (release/updates)
                "http://deb.debian.org/debian-security/",        # alternate: CDN

                "http://archive.debian.org/debian/",             # Archived Debian Releases
                "http://archive.debian.org/debian-security/",    # Archived Debian Security
                "http://archive.debian.org/debian-backports/",   # Archived Debian Backports
            ],
            "components": ["main", "contrib", "non-free", "non-free-firmware"],

            "codename": {
                "sid": {
                    "apt_keys": [APT_KEYS["Debian"]["bullseye"]["archive"], APT_KEYS["Debian"]["bookworm"]["archive"]],
                },
                "forky": {
                    "apt_keys": [APT_KEYS["Debian"]["bookworm"]["archive"], APT_KEYS["Debian"]["trixie"]["archive"]],
                    "extra_sources": [
                        {
                            "codename": "forky-security",
                            "apt_keys": [APT_KEYS["Debian"]["bookworm"]["security"], APT_KEYS["Debian"]["trixie"]["security"]],
                            "priority": 500,
                            "options": {
                                "remove_from_component": "updates/",
                            },
                        },
                        {
                            "origin": "Debian Backports",
                            "codename": "forky-backports",
                            "apt_keys": [APT_KEYS["Debian"]["bookworm"]["archive"], APT_KEYS["Debian"]["trixie"]["archive"]],
                        },
                    ],
                },
                "trixie": {
                    "apt_keys": [APT_KEYS["Debian"]["bookworm"]["archive"], APT_KEYS["Debian"]["trixie"]["archive"], APT_KEYS["Debian"]["trixie"]["release"]],
                    "extra_sources": [
                        {
                            "codename": "trixie-security",
                            "apt_keys": [APT_KEYS["Debian"]["bookworm"]["security"], APT_KEYS["Debian"]["trixie"]["security"]],
                            "priority": 500,
                            "options": {
                                "remove_from_component": "updates/",
                            },
                        },
                        {
                            "origin": "Debian Backports",
                            "codename": "trixie-backports",
                            "apt_keys": [APT_KEYS["Debian"]["bookworm"]["archive"], APT_KEYS["Debian"]["trixie"]["archive"]],
                        },
                    ],
                },
                "bookworm": {
                    "lintian_version": "2.116.3",
                    "apt_keys": [APT_KEYS["Debian"]["buster"]["archive"], APT_KEYS["Debian"]["bullseye"]["archive"], APT_KEYS["Debian"]["bookworm"]["release"]],
                    "extra_sources": [
                        {
                            "codename": "bookworm-security",
                            "apt_keys": [APT_KEYS["Debian"]["bullseye"]["security"], APT_KEYS["Debian"]["bookworm"]["security"]],
                            "priority": 500,
                            "options": {
                                "remove_from_component": "updates/",
                            },
                        },
                        {
                            "origin": "Debian Backports",
                            "codename": "bookworm-backports",
                            "apt_keys": [APT_KEYS["Debian"]["bullseye"]["archive"], APT_KEYS["Debian"]["bookworm"]["archive"]],
                        },
                        {
                            "origin": "Debian Backports",
                            "codename": "bookworm-backports-sloppy",
                            "apt_keys": [APT_KEYS["Debian"]["bullseye"]["archive"], APT_KEYS["Debian"]["bookworm"]["archive"]],
                        },
                    ],
                },
                "bullseye": {
                    "lintian_version": "2.104.0",
                    "components": ["main", "contrib", "non-free"],
                    "apt_keys": [APT_KEYS["Debian"]["buster"]["archive"], APT_KEYS["Debian"]["bullseye"]["archive"], APT_KEYS["Debian"]["bullseye"]["release"]],
                    "extra_sources": [
                        {
                            "codename": "bullseye-security",
                            "apt_keys": [APT_KEYS["Debian"]["buster"]["security"], APT_KEYS["Debian"]["bullseye"]["security"]],
                            "priority": 500,
                            "options": {
                                "remove_from_component": "updates/",
                            },
                        },
                        {
                            "origin": "Debian Backports",
                            "codename": "bullseye-backports",
                            "apt_keys": [APT_KEYS["Debian"]["buster"]["archive"], APT_KEYS["Debian"]["bullseye"]["archive"], APT_KEYS["Debian"]["bookworm"]["archive"]],
                        },
                        {
                            "origin": "Debian Backports",
                            "codename": "bullseye-backports-sloppy",
                            "apt_keys": [APT_KEYS["Debian"]["buster"]["archive"], APT_KEYS["Debian"]["bullseye"]["archive"], APT_KEYS["Debian"]["bookworm"]["archive"]],
                        },
                    ],
                },
                "buster": {
                    "lintian_version": "2.15.0",
                    "components": ["main", "contrib", "non-free"],
                    "apt_keys": [APT_KEYS["Debian"]["buster"]["archive"], APT_KEYS["Debian"]["buster"]["release"], APT_KEYS["Debian"]["bullseye"]["archive"]],
                    "extra_sources": [
                        {
                            "codename": "buster/updates",
                            "apt_keys": [APT_KEYS["Debian"]["buster"]["security"], APT_KEYS["Debian"]["bullseye"]["security"]],
                            "priority": 500,
                            "options": {
                                "extra_identifiers": {
                                    "Codename": "buster",
                                    "Label": "Debian-Security",
                                },
                                "remove_from_component": "updates/",
                            },
                        },
                        {
                            "origin": "Debian Backports",
                            "codename": "buster-backports",
                            "apt_keys": [APT_KEYS["Debian"]["buster"]["archive"], APT_KEYS["Debian"]["bullseye"]["archive"], APT_KEYS["Debian"]["bookworm"]["archive"]],
                        },
                        {
                            "origin": "Debian Backports",
                            "codename": "buster-backports-sloppy",
                            "apt_keys": [APT_KEYS["Debian"]["buster"]["archive"], APT_KEYS["Debian"]["bullseye"]["archive"], APT_KEYS["Debian"]["bookworm"]["archive"]],
                        },
                    ],
                    "distribution": {
                        "options": {
                            "lintian_extra_options": "--info --suppress-tags bad-distribution-in-changes-file",
                            "lintian_warnfail_options": "--fail-on-warnings",
                        },
                    },
                },
                "stretch": {
                    "lintian_version": "2.5.50.4",
                    "components": ["main", "contrib", "non-free"],
                    "apt_keys": [APT_KEYS["Debian"]["stretch"]["archive"], APT_KEYS["Debian"]["buster"]["archive"], APT_KEYS["Debian"]["bullseye"]["archive"], APT_KEYS["Debian"]["stretch"]["release"]],
                    "extra_sources": [
                        {
                            "codename": "stretch/updates",
                            "apt_keys": [APT_KEYS["Debian"]["stretch"]["security"], APT_KEYS["Debian"]["buster"]["security"]],
                            "priority": 500,
                            "options": {
                                "extra_identifiers": {
                                    "Codename": "stretch",
                                    "Label": "Debian-Security",
                                },
                                "remove_from_component": "updates/",
                            },
                        },
                        {
                            "origin": "Debian Backports",
                            "codename": "stretch-backports",
                            "apt_keys": [APT_KEYS["Debian"]["buster"]["archive"], APT_KEYS["Debian"]["bullseye"]["archive"]],
                        },
                        {
                            "origin": "Debian Backports",
                            "codename": "stretch-backports-sloppy",
                            "apt_keys": [APT_KEYS["Debian"]["buster"]["archive"], APT_KEYS["Debian"]["bullseye"]["archive"]],
                        },
                    ],
                    "distribution": {
                        "options": {
                            "lintian_extra_options": "--info --suppress-tags bad-distribution-in-changes-file",
                            "lintian_warnfail_options": "--fail-on-warnings",
                        },
                    },
                },
                "jessie": {
                    "lintian_version": "2.5.30+deb8u4",
                    "components": ["main", "contrib", "non-free"],
                    "apt_keys": [APT_KEYS["Debian"]["wheezy"]["archive"], APT_KEYS["Debian"]["jessie"]["release"], APT_KEYS["Debian"]["jessie"]["archive"]],
                    "extra_sources": [
                        {
                            "codename": "jessie/updates",
                            "apt_keys": [APT_KEYS["Debian"]["jessie"]["security"], APT_KEYS["Debian"]["stretch"]["security"]],
                            "priority": 500,
                            "options": {
                                "extra_identifiers": {
                                    "Codename": "jessie",
                                    "Label": "Debian-Security",
                                },
                                "remove_from_component": "updates/",
                            },
                        },
                        {
                            "origin": "Debian Backports",
                            "codename": "jessie-backports",
                            "apt_keys": [APT_KEYS["Debian"]["wheezy"]["archive"], APT_KEYS["Debian"]["jessie"]["archive"]],
                            "options": {
                                "check_valid_until": False,
                            },
                        },
                        {
                            "origin": "Debian Backports",
                            "codename": "jessie-backports-sloppy",
                            "apt_keys": [APT_KEYS["Debian"]["wheezy"]["archive"], APT_KEYS["Debian"]["jessie"]["archive"]],
                            "options": {
                                "check_valid_until": False,
                            },
                        },
                    ],
                    "distribution": {
                        "options": {
                            # jessie sources 'archive.debian.org' and 'ftp.debian.org' are not the same:
                            # * jessie from archive.debian.org needs this setting (as the Release file is signed with an expired key, and jessie's apt just won't work).
                            # * jessie from ftp.debian.org (was left there for LTS for 'amd64 armel armhf i386' only) would work w/o that flag, as it is signed differently.
                            # However, both archives will match the jessie source, and both could potentially be used.
                            "sbuild_setup_blocks": "apt-allow-unauthenticated apt-setup apt-key-binary apt-update ccache eatmydata",
                            "lintian_extra_options": "--info --suppress-tags bad-distribution-in-changes-file",
                            "lintian_warnfail_options": "--fail-on-warnings",
                        },
                    },
                },
                "wheezy": {
                    "lintian_version": "2.5.10.4",
                    "components": ["main", "contrib", "non-free"],
                    "apt_keys": [APT_KEYS["Debian"]["wheezy"]["archive"], APT_KEYS["Debian"]["wheezy"]["release"], APT_KEYS["Debian"]["jessie"]["archive"]],
                    "extra_sources": [
                        {
                            "codename": "wheezy/updates",
                            "apt_keys": [APT_KEYS["Debian"]["wheezy"]["archive"], APT_KEYS["Debian"]["jessie"]["security"]],
                            "priority": 500,
                            "options": {
                                "extra_identifiers": {
                                    "Codename": "wheezy",
                                    "Label": "Debian-Security",
                                },
                                "remove_from_component": "updates/",
                                "check_valid_until": False,
                            },
                        },
                        {
                            "origin": "Debian Backports",
                            "codename": "wheezy-backports",
                            "apt_keys": [APT_KEYS["Debian"]["wheezy"]["archive"], APT_KEYS["Debian"]["jessie"]["archive"]],
                        },
                        {
                            "origin": "Debian Backports",
                            "codename": "wheezy-backports-sloppy",
                            "apt_keys": [APT_KEYS["Debian"]["wheezy"]["archive"], APT_KEYS["Debian"]["jessie"]["archive"]],
                        },
                    ],
                    "distribution": {
                        "options": {
                            "sbuild_setup_blocks": "apt-allow-unauthenticated apt-setup apt-key-binary apt-update ccache eatmydata",
                            "lintian_extra_options": "--info --suppress-tags bad-distribution-in-changes-file",
                            "lintian_warnfail_options": "--fail-on-warnings",
                        },
                    },
                    "chroot": {
                        "options": {
                            "debootstrap_extra_options": "--include=realpath",
                        },
                    },
                },
                "squeeze": {
                    "broken": "Only works with sbuild < 0.84",

                    "lintian_version": "2.4.3+squeeze1",
                    "components": ["main", "contrib", "non-free"],
                    "apt_keys": [APT_KEYS["Debian"]["squeeze"]["archive"], APT_KEYS["Debian"]["squeeze"]["release"]],
                    "extra_sources": [
                        {
                            "codename": "squeeze/updates",
                            "apt_keys": [APT_KEYS["Debian"]["squeeze"]["archive"], APT_KEYS["Debian"]["jessie"]["security"]],
                            "priority": 500,
                            "options": {
                                "extra_identifiers": {
                                    "Codename": "squeeze",
                                    "Label": "Debian-Security",
                                },
                                "remove_from_component": "updates/",
                                "check_valid_until": False,
                            },
                        },
                        {
                            "codename": "squeeze-lts",
                            "apt_keys": [APT_KEYS["Debian"]["wheezy"]["archive"]],
                        },
                        {
                            "origin": "Debian Backports",
                            "codename": "squeeze-backports",
                            "apt_keys": [APT_KEYS["Debian"]["squeeze"]["archive"], APT_KEYS["Debian"]["wheezy"]["archive"]],
                        },
                    ],
                    "distribution": {
                        "options": {
                            "build_dep_resolver": 0,  # util.models().Distribution.RESOLVER_APT
                            "sbuild_setup_blocks": "apt-no-check-valid-until apt-allow-unauthenticated apt-setup apt-key-add apt-update ccache eatmydata",
                            "lintian_extra_options": "--info --suppress-tags bad-distribution-in-changes-file",
                            "lintian_warnfail_options": "--fail-on-warnings",
                        },
                    },
                    "chroot": {
                        "options": {
                            "debootstrap_extra_options": "--include=realpath",
                        },
                    },
                },
                "lenny": {
                    "broken": "Cannot build test packages (only dh7 available)",

                    "lintian_version": "1.24.2.1+lenny1",
                    "components": ["main", "contrib", "non-free"],
                    "apt_keys": [APT_KEYS["Debian"]["squeeze"]["archive"], APT_KEYS["Debian"]["lenny"]["release"]],
                    "extra_sources": [
                        {
                            "codename": "lenny/updates",
                            "apt_keys": [APT_KEYS["Debian"]["lenny"]["archive"], APT_KEYS["Debian"]["jessie"]["security"]],
                            "priority": 500,
                            "options": {
                                "extra_identifiers": {
                                    "Codename": "lenny",
                                    "Label": "Debian-Security",
                                },
                                "remove_from_component": "updates/",
                                "check_valid_until": False,
                            },
                        },
                        {
                            "origin": "Debian Backports",
                            "codename": "lenny-backports",
                            "apt_keys": [APT_KEYS["Debian"]["squeeze"]["archive"]],
                        },
                    ],
                    "distribution": {
                        "options": {
                            "sbuild_setup_blocks": "apt-no-check-valid-until apt-allow-unauthenticated apt-setup apt-key-add apt-update ccache eatmydata",
                            "lintian_extra_options": "--info",
                            "lintian_warnfail_options": "--fail-on-warnings",
                        },
                    },
                    "chroot": {
                        "options": {
                            "debootstrap_command": "/usr/sbin/mini-buildd-debootstrap-uname-2.6",
                            "debootstrap_extra_options": "--include=realpath,timeout",
                        },
                    },
                },
            },
        },
        #: Ubuntu (https://wiki.ubuntu.com/Releases)
        #:
        #: Keep all active LTS and the four most recent releases.
        "Ubuntu": {
            "archive_paths": ["ubuntu", "ubuntu-old"],
            "archive": [
                "http://archive.ubuntu.com/ubuntu/",       # Releases
                "http://security.ubuntu.com/ubuntu/",      # Security
                "http://old-releases.ubuntu.com/ubuntu/",  # Older Releases
            ],
            "components": ["main", "universe", "restricted", "multiverse"],

            "codename": {
                "plucky": {  # plucky: 25.04
                    "lintian_version": "2.121.2ubuntu1",
                    "apt_keys": [APT_KEYS["Ubuntu"]["2018"]],
                    "extra_sources": [
                        {
                            "codename": "plucky-security",
                            "priority": 500,
                            "options": {
                                "extra_identifiers": {
                                    "Codename": "plucky",
                                    "Suite": "plucky-security",
                                },
                            },
                        },
                        {
                            "codename": "plucky-backports",
                            "options": {
                                "extra_identifiers": {
                                    "Codename": "plucky",
                                    "Suite": "plucky-backports",
                                },
                            },
                        },
                    ],
                    "distribution": {
                        "options": {
                            "deb_build_options": "noddebs",
                        },
                        "arch_optional": ["i386"],
                    },
                },
                "oracular": {  # oracular: 24.10
                    "lintian_version": "2.118.0ubuntu1",
                    "apt_keys": [APT_KEYS["Ubuntu"]["2018"]],
                    "extra_sources": [
                        {
                            "codename": "oracular-security",
                            "priority": 500,
                            "options": {
                                "extra_identifiers": {
                                    "Codename": "oracular",
                                    "Suite": "oracular-security",
                                },
                            },
                        },
                        {
                            "codename": "oracular-backports",
                            "options": {
                                "extra_identifiers": {
                                    "Codename": "oracular",
                                    "Suite": "oracular-backports",
                                },
                            },
                        },
                    ],
                    "distribution": {
                        "options": {
                            "deb_build_options": "noddebs",
                        },
                        "arch_optional": ["i386"],
                    },
                },
                "noble": {  # noble: 24.04 LTS 2029
                    "lintian_version": "2.117.0ubuntu1",
                    "apt_keys": [APT_KEYS["Ubuntu"]["2018"]],
                    "extra_sources": [
                        {
                            "codename": "noble-security",
                            "priority": 500,
                            "options": {
                                "extra_identifiers": {
                                    "Codename": "noble",
                                    "Suite": "noble-security",
                                },
                            },
                        },
                        {
                            "codename": "noble-backports",
                            "options": {
                                "extra_identifiers": {
                                    "Codename": "noble",
                                    "Suite": "noble-backports",
                                },
                            },
                        },
                    ],
                    "distribution": {
                        "options": {
                            "deb_build_options": "noddebs",
                        },
                        "arch_optional": ["i386"],
                    },
                },
                "mantic": {  # mantic: 23.10
                    "lintian_version": "2.116.3ubuntu3",
                    "apt_keys": [APT_KEYS["Ubuntu"]["2018"]],
                    "extra_sources": [
                        {
                            "codename": "mantic-security",
                            "priority": 500,
                            "options": {
                                "extra_identifiers": {
                                    "Codename": "mantic",
                                    "Suite": "mantic-security",
                                },
                            },
                        },
                        {
                            "codename": "mantic-backports",
                            "options": {
                                "extra_identifiers": {
                                    "Codename": "mantic",
                                    "Suite": "mantic-backports",
                                },
                            },
                        },
                    ],
                    "distribution": {
                        "options": {
                            "deb_build_options": "noddebs",
                        },
                        "arch_optional": ["i386"],
                    },
                },
                "lunar": {  # lunar: 23.04
                    "lintian_version": "2.116.3ubuntu1",
                    "apt_keys": [APT_KEYS["Ubuntu"]["2018"]],
                    "extra_sources": [
                        {
                            "codename": "lunar-security",
                            "priority": 500,
                            "options": {
                                "extra_identifiers": {
                                    "Codename": "lunar",
                                    "Suite": "lunar-security",
                                },
                            },
                        },
                        {
                            "codename": "lunar-backports",
                            "options": {
                                "extra_identifiers": {
                                    "Codename": "lunar",
                                    "Suite": "lunar-backports",
                                },
                            },
                        },
                    ],
                    "distribution": {
                        "options": {
                            "deb_build_options": "noddebs",
                        },
                        "arch_optional": ["i386"],
                    },
                },
                "kinetic": {  # kinetic: 22.10
                    "lintian_version": "2.115.3ubuntu2",
                    "apt_keys": [APT_KEYS["Ubuntu"]["2018"]],
                    "extra_sources": [
                        {
                            "codename": "kinetic-security",
                            "priority": 500,
                            "options": {
                                "extra_identifiers": {
                                    "Codename": "kinetic",
                                    "Suite": "kinetic-security",
                                },
                            },
                        },
                        {
                            "codename": "kinetic-backports",
                            "options": {
                                "extra_identifiers": {
                                    "Codename": "kinetic",
                                    "Suite": "kinetic-backports",
                                },
                            },
                        },
                    ],
                    "distribution": {
                        "options": {
                            "deb_build_options": "noddebs",
                        },
                        "arch_optional": ["i386"],
                    },
                },
                "jammy": {  # jammy: 22.04 LTS 2027
                    "lintian_version": "2.114.0ubuntu1",
                    "apt_keys": [APT_KEYS["Ubuntu"]["2018"]],
                    "extra_sources": [
                        {
                            "codename": "jammy-security",
                            "priority": 500,
                            "options": {
                                "extra_identifiers": {
                                    "Codename": "jammy",
                                    "Suite": "jammy-security",
                                },
                            },
                        },
                        {
                            "codename": "jammy-backports",
                            "options": {
                                "extra_identifiers": {
                                    "Codename": "jammy",
                                    "Suite": "jammy-backports",
                                },
                            },
                        },
                    ],
                    "distribution": {
                        "options": {
                            "deb_build_options": "noddebs",
                        },
                        "arch_optional": ["i386"],
                    },
                },
                "impish": {  # impish: 21.10
                    "lintian_version": "2.104.0ubuntu3",
                    "apt_keys": [APT_KEYS["Ubuntu"]["2018"]],
                    "extra_sources": [
                        {
                            "codename": "impish-security",
                            "priority": 500,
                            "options": {
                                "extra_identifiers": {
                                    "Codename": "impish",
                                    "Suite": "impish-security",
                                },
                            },
                        },
                        {
                            "codename": "impish-backports",
                            "options": {
                                "extra_identifiers": {
                                    "Codename": "impish",
                                    "Suite": "impish-backports",
                                },
                            },
                        },
                    ],
                    "distribution": {
                        "options": {
                            "deb_build_options": "noddebs",
                        },
                        "arch_optional": ["i386"],
                    },
                },
                "hirsute": {  # hirsute: 21.04
                    "lintian_version": "2.104.0ubuntu1",
                    "apt_keys": [APT_KEYS["Ubuntu"]["2018"]],
                    "extra_sources": [
                        {
                            "codename": "hirsute-security",
                            "priority": 500,
                            "options": {
                                "extra_identifiers": {
                                    "Codename": "hirsute",
                                    "Suite": "hirsute-security",
                                },
                            },
                        },
                        {
                            "codename": "hirsute-backports",
                            "options": {
                                "extra_identifiers": {
                                    "Codename": "hirsute",
                                    "Suite": "hirsute-backports",
                                },
                            },
                        },
                    ],
                    "distribution": {
                        "options": {
                            "deb_build_options": "noddebs",
                            "lintian_extra_options": "--info --suppress-tags bad-distribution-in-changes-file",
                        },
                        "arch_optional": ["i386"],
                    },
                },
                "groovy": {  # groovy: 20.10
                    "lintian_version": "2.89.0",
                    "apt_keys": [APT_KEYS["Ubuntu"]["2018"]],
                    "extra_sources": [
                        {
                            "codename": "groovy-security",
                            "priority": 500,
                            "options": {
                                "extra_identifiers": {
                                    "Codename": "groovy",
                                    "Suite": "groovy-security",
                                },
                            },
                        },
                        {
                            "codename": "groovy-backports",
                            "options": {
                                "extra_identifiers": {
                                    "Codename": "groovy",
                                    "Suite": "groovy-backports",
                                },
                            },
                        },
                    ],
                    "distribution": {
                        "options": {
                            "deb_build_options": "noddebs",
                            "lintian_extra_options": "--info --suppress-tags bad-distribution-in-changes-file",
                        },
                        "arch_optional": ["i386"],
                    },
                },
                "focal": {  # focal: 20.04 LTS 2025 ESM 2030
                    "lintian_version": "2.62.0",
                    "apt_keys": [APT_KEYS["Ubuntu"]["2012"], APT_KEYS["Ubuntu"]["2018"]],
                    "extra_sources": [
                        {
                            "codename": "focal-security",
                            "priority": 500,
                            "options": {
                                "extra_identifiers": {
                                    "Codename": "focal",
                                    "Suite": "focal-security",
                                },
                            },
                        },
                        {
                            "codename": "focal-backports",
                            "options": {
                                "extra_identifiers": {
                                    "Codename": "focal",
                                    "Suite": "focal-backports",
                                },
                            },
                        },
                    ],
                    "distribution": {
                        "options": {
                            "deb_build_options": "noddebs",
                            "lintian_extra_options": "--info --suppress-tags bad-distribution-in-changes-file",
                        },
                        # Ubuntu drops i386 (starting with focal)
                        # 'focal' still has i386 repo (debootrap still works), so i386 is still picked
                        # by mini-buildd's setup. However, builds no longer work (default resolver
                        # aptitude missing).
                        # Choosing "apt" resolver would master this hurdle, but this would also change
                        # the default resolver for amd64. Imho best solution for mini-buildd's setup
                        # is to make this arch optional.
                        # Ref: https://discourse.ubuntu.com/t/community-process-for-32-bit-compatibility/12598?u=d0od
                        "arch_optional": ["i386"],
                        "lintian_warnfail_options": "",
                    },
                },
                "bionic": {  # bionic: 18.04 LTS 2023 ESM 2028
                    "lintian_version": "2.5.81ubuntu1",
                    "apt_keys": [APT_KEYS["Ubuntu"]["2012"]],
                    "extra_sources": [
                        {
                            "codename": "bionic-security",
                            "priority": 500,
                            "options": {
                                "extra_identifiers": {
                                    "Codename": "bionic",
                                    "Suite": "bionic-security",
                                },
                            },
                        },
                        {
                            "codename": "bionic-backports",
                            "options": {
                                "extra_identifiers": {
                                    "Codename": "bionic",
                                    "Suite": "bionic-backports",
                                },
                            },
                        },
                    ],
                    "distribution": {
                        "options": {
                            # Starting with cosmic (18.10), Ubuntu uses 'ddeb' as file appendix for automated debug packages.
                            # reprepro can't handle these yet -- so this is needed for a workaround.
                            "deb_build_options": "noddebs",
                            "lintian_extra_options": "--info --suppress-tags bad-distribution-in-changes-file",
                            "lintian_warnfail_options": "--fail-on-warnings",
                        },
                    },
                },
                "xenial": {  # xenial: 16.04 LTS 2021 ESM 2026
                    "lintian_version": "2.5.43",
                    "apt_keys": [APT_KEYS["Ubuntu"]["2004"], APT_KEYS["Ubuntu"]["2012"]],
                    "extra_sources": [
                        {
                            "codename": "xenial-security",
                            "priority": 500,
                            "options": {
                                "extra_identifiers": {
                                    "Codename": "xenial",
                                    "Suite": "xenial-security",
                                },
                            },
                        },
                        {
                            "codename": "xenial-backports",
                            "options": {
                                "extra_identifiers": {
                                    "Codename": "xenial",
                                    "Suite": "xenial-backports",
                                },
                            },
                        },
                    ],
                    "distribution": {
                        "options": {
                            "sbuild_setup_blocks": "apt-setup apt-key-add apt-update ccache eatmydata",
                            "lintian_extra_options": "--info --suppress-tags bad-distribution-in-changes-file",
                            "lintian_warnfail_options": "--fail-on-warnings",
                        },
                    },
                },
                "trusty": {  # trusty: 14.04 LTS 2019 ESM 2024
                    "broken": "``upstart`` install fails in build chroots",

                    "lintian_version": "2.5.22ubuntu1",
                    "apt_keys": [APT_KEYS["Ubuntu"]["2004"], APT_KEYS["Ubuntu"]["2012"]],
                    "extra_sources": [
                        {
                            "codename": "trusty-security",
                            "priority": 500,
                            "options": {
                                "extra_identifiers": {
                                    "Codename": "trusty",
                                    "Suite": "trusty-security",
                                },
                            },
                        },
                        {
                            "codename": "trusty-backports",
                            "options": {
                                "extra_identifiers": {
                                    "Codename": "trusty",
                                    "Suite": "trusty-backports",
                                },
                            },
                        },
                    ],
                    "distribution": {
                        "options": {
                            "sbuild_setup_blocks": "apt-setup apt-key-add apt-update ccache eatmydata",
                            "lintian_extra_options": "--info --suppress-tags bad-distribution-in-changes-file",
                            "lintian_warnfail_options": "--fail-on-warnings",
                        },
                    },
                    "chroot": {
                        "options": {
                            "debootstrap_extra_options": "--include=realpath",
                        },
                    },
                },
            },
        },

        #: PureOS (https://pureos.net/)
        "PureOS": {
            "archive_paths": [],
            "archive": [
                "https://mirror.fsf.org/pureos/",
            ],
            "components": ["main"],

            "codename": {
                "amber": {
                    "apt_keys": [APT_KEYS["PureOS"]["repository"]],
                    "extra_sources": [
                        {
                            "codename": "amber-security",
                            "priority": 500,
                        },
                        {
                            "codename": "amber-updates",
                        }
                    ],
                },
            },
        },
    },
}


def setup_origins():
    return SETUP["origin"].keys()


def setup_codenames():
    codenames = []
    for o in setup_origins():
        codenames.extend(SETUP["origin"][o]["codename"])
    return codenames


VENDOR_CODENAMES_FUNC = {
    "": lambda origin: [codename for codename in DISTRO_INFO[origin].supported() if codename in setup_codenames()],
    "lts": lambda origin: [codename for codename in DISTRO_INFO[origin].supported() + DISTRO_INFO[origin].mbd_lts() if codename in setup_codenames()],
    "all": lambda origin: [codename for codename, setup in SETUP["origin"][origin]["codename"].items() if setup.get("broken") is None],
}


def setup_codenames_from_origin(origin, modifier=""):
    try:
        return VENDOR_CODENAMES_FUNC[modifier](origin)
    except KeyError as e:
        raise util.HTTPBadRequest(f"Unknown codenames selector: '{origin}:{modifier}' (use '{'|'.join(setup_origins())}[:{'|'.join(VENDOR_CODENAMES_FUNC.keys())}]'") from e


def _idi_codenames():
    for di in DISTRO_INFO.values():
        yield from di.all


def di_codenames():
    return list(_idi_codenames())


def guess_origin(codename):
    """
    Guess Origin from codename

    >>> guess_origin("rex")  # only via distro info
    'Debian'
    >>> guess_origin("warty")  # only via distro info
    'Ubuntu'
    >>> guess_origin("bullseye")  # via setup
    'Debian'
    """
    for origin, setup in SETUP["origin"].items():
        if codename in setup["codename"]:
            return origin

        # Check if we can find codename in distro info
        di = DISTRO_INFO.get(origin)
        if di and codename in di.all:
            return origin
    return None


def is_base_source(origin, codename):
    """Heuristics (for setup) whether this is a base source"""
    return origin in SETUP["origin"] and re.match(r"^[a-z]+$", codename) and codename != "experimental"


#: List of Debian codenames using something ``3.1`` as main version (all <= squeeze)
DEBIAN_MAJOR_MINOR_CODENAMES = ["buzz", "rex", "bo", "hamm", "slink", "potato", "woody", "sarge", "etch", "lenny", "squeeze"]


def guess_codeversion(origin, codename, version):
    """
    Get recommended codeversion from origin/codename/version triple

    ``codename`` is essentially one number representing the release version, which can later be used as in
    mandatory version part, or for sorting codenames.

    ``version`` may be from a ``Release`` file (like '11.3' for bullseye), or from ``distro-info-data``
    (just '11'). If empty, ``~CODENAME`` is returned.

    Some heuristics (tailored for Debian and Ubuntu, but may work fine for other origins) are applied to
    produce a reasonable ``codeversion``.

    In Debian,
      - point release <= sarge had the 'M.PrN' syntax (with 3.1 being a major release).
      - point release in squeeze used 'M.0.N' syntax.
      - point releases for >= wheezy have the 'M.N' syntax (with 7.1 being a point release).
      - testing and unstable do not gave a version in Release and fall back to uppercase codename

    Ubuntu just uses YY.MM which we can use as-is.

    >>> guess_codeversion("Debian", "sarge", "3.1r8")
    '31'
    >>> guess_codeversion("Debian", "etch", "4.0r9")
    '40'
    >>> guess_codeversion("Debian", "squeeze", "6.0.6")
    '60'
    >>> guess_codeversion("Debian", "wheezy", "7.0")
    '7'
    >>> guess_codeversion("Debian", "wheezy", "7.1")
    '7'
    >>> guess_codeversion("Ubuntu", "quantal", "12.10")
    '1210'
    """
    if not version:
        return "~" + codename.upper()

    ver_split = version.split(".")

    def number(no):
        """Be sure to not fail (return empty str instead), and only use *starting* digits"""
        v = ver_split[no] if len(ver_split) > no else ""
        m = re.search(r"^(\d+)", v)
        return m.group(1) if m else ""

    if origin == "Debian" and codename not in DEBIAN_MAJOR_MINOR_CODENAMES:
        return number(0)  # Debian >= wheezy
    return f"{number(0)}{number(1)}"


@functools.total_ordering
class Codename():
    def __init__(self, codename, origin="", version=""):
        self.codename = codename
        self.origin = origin if origin else guess_origin(codename)

        di = DISTRO_INFO.get(self.origin)

        # Set or compute version (may be empty string)
        if not version and di is not None:
            r = di.mbd_release(codename)
            version = di.version(self.codename) if (r is not None and r.release is not None) else ""  # Version might be non-empty in distro-info-data even if not released

        # Guess codeversion
        self.codeversion = guess_codeversion(self.origin, self.codename, version)

        # Compute cmpversion
        try:
            if di is None:
                raise util.HTTPUnavailable("No distro info")

            version = di.version(codename)
            if not version:
                raise util.HTTPUnavailable("No version in distro info")

            self.cmpversion = debian.debian_support.Version(version.replace(" ", "+"))  # Some Ubuntu versions need " LTS"->"+LTS" to produce valid version strings
        except Exception:
            self.cmpversion = debian.debian_support.Version("9999")

        self.setup = SETUP["origin"].get(self.origin, {}).get("codename", {}).get(self.codename)
        self.has_setup = self.setup is not None
        if not self.has_setup:
            self.setup = self._default_setup()

    def _default_setup(self):
        if self.origin == "Debian":
            return {
                "extra_sources": [
                    {
                        "codename": f"{self.codename}-security",
                        "priority": 500,
                        "options": {
                            "remove_from_component": "updates/",
                        },
                    },
                    {
                        "origin": "Debian Backports",
                        "codename": f"{self.codename}-backports",
                    },
                ]
            }
        if self.origin == "Ubuntu":
            return {
                "apt_keys": [APT_KEYS["Ubuntu"]["2018"]],
                "extra_sources": [
                    {
                        "codename": f"{self.codename}-security",
                        "priority": 500,
                        "options": {
                            "extra_identifiers": {"Codename": self.codename, "Suite": f"{self.codename}-security"},
                        },
                    },
                    {
                        "codename": f"{self.codename}-backports",
                        "options": {
                            "extra_identifiers": {"Codename": self.codename, "Suite": f"{self.codename}-backports"},
                        },
                    },
                ],
                "distribution": {
                    "options": {
                        "deb_build_options": "noddebs",
                    },
                    "arch_optional": ["i386"],
                }
            }
        return {}

    def setup_extra_sources(self):
        return self.setup.get("extra_sources", [])

    def setup_distribution_options(self):
        return util.mdget(self.setup, ["distribution", "options"], {})

    def setup_chroot_options(self):
        return util.mdget(self.setup, ["chroot", "options"], {})

    def setup_arch_optional(self):
        return util.mdget(self.setup, ["distribution", "arch_optional"], [])

    def setup_components(self):
        return util.mdget(self.setup, ["components"], util.mdget(SETUP, ["origin", self.origin, "components"], ["main"]))

    def __str__(self):
        return f"{self.origin} '{self.codename}' ({self.codeversion})"

    def __hash__(self):
        return hash(self.codename)

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.origin == other.origin and self.codename == other.codename
        return NotImplemented

    def __lt__(self, other):
        if isinstance(other, self.__class__):
            if self.origin == other.origin:
                return self.cmpversion < other.cmpversion
            return self.origin and other.origin and self.origin < other.origin
        return NotImplemented


class SbuildCheck():
    """
    Generic support for sbuild checks (lintian, piuparts, autopkgtest)

    >>> SbuildCheck("lindian", "disabled")  # doctest: +IGNORE_EXCEPTION_DETAIL
    Traceback (most recent call last):
    ...
    util.HTTPBadRequest: Unknown sbuild checker: lindian (valid options: lintian,piuparts,autopkgtest) (HTTP 400 Bad request syntax or unsupported method)

    >>> SbuildCheck("lintian", "warnfall")  # doctest: +IGNORE_EXCEPTION_DETAIL
    Traceback (most recent call last):
    ...
    util.HTTPBadRequest: Unknown sbuild check mode: warnfall (valid options: DISABLED,IGNORE,ERRFAIL,WARNFAIL) (HTTP 400 Bad request syntax or unsupported method)

    >>> sc = SbuildCheck("lintian", "warnfail")
    >>> sc.checker
    'lintian'
    >>> sc.mode
    <Mode.WARNFAIL: 3>
    """

    class Mode(enum.Enum):
        DISABLED = 0
        IGNORE = 1
        ERRFAIL = 2
        WARNFAIL = 3

        def __str__(self):
            return {
                self.DISABLED: "Don't run check",
                self.IGNORE: "Run check but ignore results",
                self.ERRFAIL: "Run check and fail on errors",
                self.WARNFAIL: "Run check and fail on warnings",
            }[self]

    CHECKERS = ["lintian", "piuparts", "autopkgtest"]
    CHOICES = [(mode.value, mode.name) for mode in Mode]

    #: From sbuild source code: We may expect these textual statuses:
    STATUSES_PASS = ["pass", "info", "successful"]
    STATUSES_WARN = ["warn", "no tests", "unknown"]
    STATUSES_FAIL = ["error", "fail", "failed"]

    def __init__(self, checker, mode):
        if checker not in self.CHECKERS:
            raise util.HTTPBadRequest(f"Unknown sbuild checker: {checker} (valid options: {','.join(self.CHECKERS)})")
        self.checker = checker
        try:
            self.mode = self.Mode[mode.upper() if isinstance(mode, str) else mode]
        except KeyError as e:
            raise util.HTTPBadRequest(f"Unknown sbuild check mode: {mode} (valid options: {','.join([mode.name for mode in self.Mode])})") from e

    @classmethod
    def desc(cls):
        return "Mode to control if a check should prevent package installation (for non-experimental suites)."

    @classmethod
    def usage(cls):
        return ", ".join([f"'{mode.name}' ({mode})" for mode in cls.Mode])

    def check(self, status, ignore=False):
        """Check if status is ok in this mode"""
        return \
            ignore or \
            status in self.STATUSES_PASS or \
            self.mode in [self.Mode.DISABLED, self.Mode.IGNORE] or \
            (status in self.STATUSES_WARN and self.mode is not self.Mode.WARNFAIL)


class Dist():
    """
    A mini-buildd distribution string

    Normal distribution:

    >>> d = Dist("squeeze-test-stable")
    >>> d.codename, d.repository, d.suite
    ('squeeze', 'test', 'stable')
    >>> d.get()
    'squeeze-test-stable'

    Rollback distribution:

    >>> d = Dist("squeeze-test-stable-rollback5")
    >>> d.is_rollback
    True
    >>> d.get(rollback=False)
    'squeeze-test-stable'
    >>> d.codename, d.repository, d.suite, d.rollback
    ('squeeze', 'test', 'stable', 'rollback5')
    >>> d.get()
    'squeeze-test-stable-rollback5'
    >>> d.rollback_no
    5

    Malformed distributions:

    >>> Dist("-squeeze-stable")  # doctest: +IGNORE_EXCEPTION_DETAIL
    Traceback (most recent call last):
    ...
    util.HTTPBadRequest: Malformed distribution '-squeeze-stable': Must be '<codename>-<repoid>-<suite>[-rollback<n>]' (HTTP 400 Bad request syntax or unsupported method)

    >>> Dist("squeeze--stable")  # doctest: +IGNORE_EXCEPTION_DETAIL
    Traceback (most recent call last):
    ...
    util.HTTPBadRequest: Malformed distribution 'squeeze--stable': Must be '<codename>-<repoid>-<suite>[-rollback<n>]' (HTTP 400 Bad request syntax or unsupported method)

    >>> Dist("squeeze-test-stable-")  # doctest: +IGNORE_EXCEPTION_DETAIL
    Traceback (most recent call last):
    ...
    util.HTTPBadRequest: Malformed distribution 'squeeze-test-stable-': Must be '<codename>-<repoid>-<suite>[-rollback<n>]' (HTTP 400 Bad request syntax or unsupported method)

    >>> Dist("squeeze-test-stable-rollback")  # doctest: +IGNORE_EXCEPTION_DETAIL
    Traceback (most recent call last):
    ...
    util.HTTPBadRequest: Malformed distribution 'squeeze-test-stable-rollback': Must be '<codename>-<repoid>-<suite>[-rollback<n>]' (HTTP 400 Bad request syntax or unsupported method)

    >>> Dist("squeeze-test-stable-rolback0")  # doctest: +IGNORE_EXCEPTION_DETAIL
    Traceback (most recent call last):
    ...
    util.HTTPBadRequest: Malformed distribution 'squeeze-test-stable-rolback0': Must be '<codename>-<repoid>-<suite>[-rollback<n>]' (HTTP 400 Bad request syntax or unsupported method)
    """

    _REGEX = re.compile(r"^\w+-\w+-\w+?(-rollback\d+)?$")

    def __init__(self, diststr):
        if not self._REGEX.match(diststr):
            raise util.HTTPBadRequest(f"Malformed distribution '{diststr}': Must be '<codename>-<repoid>-<suite>[-rollback<n>]'")

        self.diststr = diststr
        self._dsplit = self.diststr.split("-")
        self.codename = self._dsplit[0]
        self.repository = self._dsplit[1]
        self.suite = self._dsplit[2]
        self.is_rollback = len(self._dsplit) == 4
        self.rollback = self._dsplit[3] if self.is_rollback else None
        self.rollback_no = int(re.sub(r"\D", "", self.rollback)) if self.rollback else None

    def get(self, rollback=True):
        return "-".join(self._dsplit) if rollback else "-".join(self._dsplit[:3])


if __name__ == "__main__":
    for DI in [DebianDistroInfo(), UbuntuDistroInfo()]:
        for C in DI.all:
            print(Codename(C))
    print(setup_origins())
    print(setup_codenames())
    print(setup_codenames_from_origin("Debian"))
    print(setup_codenames_from_origin("Debian", "all"))
    print(setup_codenames_from_origin("Debian", "lts"))
    print(di_codenames())

    C = Codename("jdaskljk")
    CL = [Codename(C) for C in DebianDistroInfo().all]
    print([f"{C.codename} {C.cmpversion}" for C in sorted(CL + [C])])
    CL = [Codename(C) for C in UbuntuDistroInfo().all]
    print([f"{C.codename} {C.cmpversion}" for C in sorted(CL + [C])])



import datetime
import json
import logging
import os.path

import twisted.internet.endpoints
import twisted.internet.threads
import twisted.python.log
import twisted.python.logfile
import twisted.python.threadpool
import twisted.web.resource
import twisted.web.server
import twisted.web.static
import twisted.web.wsgi
from twisted.internet import reactor

from mini_buildd import config, daemon, events, net, threads, util

LOG = logging.getLogger(__name__)


class Site(twisted.web.server.Site):
    def _openLogFile(self, path):  # noqa (pep8 N802)
        return twisted.python.logfile.LogFile(os.path.basename(path), directory=os.path.dirname(path), rotateLength=5000000, maxRotatedFiles=9)


class RootResource(twisted.web.resource.Resource):
    """Twisted root resource needed to mix static and wsgi resources"""

    def __init__(self, wsgi_resource):
        super().__init__()
        self._wsgi_resource = wsgi_resource

    def getChild(self, path, request):  # noqa (pep8 N802)
        request.prepath.pop()
        request.postpath.insert(0, path)
        return self._wsgi_resource


class Events(twisted.web.resource.Resource):
    @classmethod
    def _render(cls, request):
        LOG.debug("%s: Waiting for event...", request.channel)
        after = datetime.datetime.fromisoformat(request.args[b"after"][0].decode(config.CHAR_ENCODING)) if b"after" in request.args else None
        queue = daemon.get().events.attach(request.channel, after=util.Datetime.check_aware(after))
        event = events.get(queue)
        LOG.debug("%s: Got event: %s", request.channel, event)
        request.write(json.dumps(event.to_json()).encode(config.CHAR_ENCODING))
        request.finish()

    @classmethod
    def _on_error(cls, failure, request):
        http_exception = util.e2http(failure.value)
        request.setResponseCode(http_exception.status, f"{http_exception}".encode(config.CHAR_ENCODING))
        if getattr(request, "_disconnected", False):
            request.notifyFinish()
        else:
            request.finish()

    def render_GET(self, request):  # noqa (pep8 N802)
        request.setHeader("Content-Type", f"application/json; charset={config.CHAR_ENCODING}")
        twisted.internet.threads.deferToThread(self._render, request).addErrback(self._on_error, request)
        return twisted.web.server.NOT_DONE_YET


class HttpD(threads.Thread):
    def _add_route(self, uri, resource):
        """Add route from (possibly nested) path from config -- making sure already existing parent routes are reused"""
        LOG.debug("Adding route: %s -> %s", uri, resource)

        uri_split = uri.twisted().split("/")
        res = self.resource
        for u in [uri_split[0:p] for p in range(1, len(uri_split))]:
            path = "/".join(u)
            sub = self.route_hierarchy.get(path)
            if sub is None:
                sub = twisted.web.resource.Resource()
                res.putChild(bytes(u[-1], encoding=config.CHAR_ENCODING), sub)
                self.route_hierarchy[path] = sub
            res = sub
        res.putChild(bytes(uri_split[-1], encoding=config.CHAR_ENCODING), resource)

    def __init__(self, wsgi_app, minthreads=0, maxthreads=10):
        # Use python's native logging system
        twisted.python.log.PythonLoggingObserver().start()

        self.endpoints = [net.ServerEndpoint(ep, protocol="http") for ep in config.HTTP_ENDPOINTS]
        super().__init__()
        self.route_hierarchy = {}

        # HTTP setup
        self.thread_pool = twisted.python.threadpool.ThreadPool(minthreads, maxthreads, name=self.__class__.__name__)
        self.resource = RootResource(twisted.web.wsgi.WSGIResource(reactor, self.thread_pool, wsgi_app))
        self.site = Site(self.resource, logPath=config.ROUTES["log"].path.join(config.ACCESS_LOG_FILE))

        # Static routes
        for route in config.ROUTES.values():
            for uri in (uri for key, uri in route.uris.items() if key.startswith("static")):
                self._add_route(uri, twisted.web.static.File(route.path.full, defaultType=f"text/plain; charset={config.CHAR_ENCODING}"))

        # Events route
        self._add_route(config.ROUTES["events"].uris["attach"], Events())

        # Start sockets
        ep_errors = []

        def on_ep_error(f):
            ep_errors.append(f.value)

        for ep in self.endpoints:
            twisted.internet.endpoints.serverFromString(reactor, ep.description).listen(self.site).addErrback(on_ep_error)

        for e in ep_errors:
            raise util.HTTPUnavailable(str(e))

    def __str__(self):
        return f"{super().__str__()} ({','.join([ep.geturl() for ep in self.endpoints])})"

    def shutdown(self):
        self.thread_pool.stop()
        reactor.stop()

    def mbd_run(self):
        self.thread_pool.start()
        reactor.run(installSignalHandlers=0)



# Generated by Django 4.2.14 on 2024-08-03 13:06

from django.db import migrations

# Previously, a hardcoded default was used in case the fields were left blank.
# We now need to fill in these hardcoded defaults.
SBUILD_CONFIG_BLOCKS_DEFAULT = "set-default-path ccache"
SBUILD_SETUP_BLOCKS_DEFAULT = "apt-setup apt-update ccache eatmydata"


def migrate_sbuild_blocks(apps, _schema_editor):
    Distribution = apps.get_model("mini_buildd", "Distribution")
    for d in Distribution.objects.all():
        if not d.sbuild_config_blocks:
            d.sbuild_config_blocks = SBUILD_CONFIG_BLOCKS_DEFAULT
        if not d.sbuild_setup_blocks:
            d.sbuild_setup_blocks = SBUILD_SETUP_BLOCKS_DEFAULT
        d.save()


class Migration(migrations.Migration):

    dependencies = [
        ('mini_buildd', '0035_alter_distribution_sbuild_config_blocks_and_more'),
    ]

    operations = [
        migrations.RunPython(migrate_sbuild_blocks),
    ]



import fnmatch
import logging
import os
import pathlib

from mini_buildd import files, util

LOG = logging.getLogger(__name__)


class Dir:
    def __init__(self, basedir, pooldir="pool", sources_glob="*/*/*"):
        self.path = pathlib.Path(basedir, pooldir)
        self.sources_glob = sources_glob

    BINARY_SUFFIXES = [".deb", ".udeb", ".ddeb"]

    def isearch(self, pattern):
        """Search pattern in all source or binary package names, return matching source package names"""
        for s in self.path.glob(self.sources_glob):
            if fnmatch.fnmatch(s.name, pattern):
                yield s.name, s
            for f in (f for f in s.rglob("*") if f.suffix in self.BINARY_SUFFIXES):
                binary = f.name.partition("_")[0]
                if fnmatch.fnmatch(binary, pattern):
                    yield s.name, f

    def dsc_path(self, source, version, raise_exception=True):
        """Get DSC pool path of an installed source (``<repo>/pool/...``)"""
        dscs = list(self.path.glob(os.path.join("*", "*", source, files.DebianName(source, version).dsc())))

        LOG.debug("Pool %s: Found pool DSCs for '%s-%s': %s", self.path, source, version, dscs)

        if len(dscs) > 1:  # This should not really ever happen. Pool-wise it could however with different components (main, contrib, etc)
            LOG.warning("Found multiple DSCs for '%s-%s' in pool (only the 1st found will be used)", source, version)
            for d in dscs:
                LOG.warning("↳ %s", d)

        if len(dscs) < 1:
            msg = f"Can't find DSC for '{source}-{version}' in pool '{self.path}'"
            if raise_exception:
                raise util.HTTPBadRequest(msg)
            LOG.warning(msg)
            return None

        return str(dscs[0].relative_to(self.path.parent.parent))


if __name__ == "__main__":
    pool = Dir("/var/lib/mini-buildd/repositories/test")
    print(set(pool.isearch("*")))



import http
import logging
import os
import pathlib

import django.contrib.auth
import django.forms
import django.http
import django.shortcuts
import django.template.loader
import django.template.response
import django.utils
import django.views
from django.contrib.auth.decorators import login_required, user_passes_test
# Decorator shortcuts
from django.utils.decorators import method_decorator
from django.views.decorators.http import require_GET

from mini_buildd import api, config, daemon, events, models, util

LOG = logging.getLogger(__name__)

# We will use custom class variables on class based views instead of extra context:
#  The view object is always automatically available in context, and this is way more
#  simple, straightforward and works well with subclassing. Also, kwargs from urls.py
#  will then be automatically available in context.
# For a django view class, this needs to be set in setup(), not __init__():
# pylint: disable=attribute-defined-outside-init


def context(_request):
    """Generate generic context. Will be available in any render, see ``django_settings.py``"""
    return {
        "mbd": {
            "VERSIONS": util.VERSIONS,
            "models": util.models,
            "config": config,
            "api": api,
            "daemon": daemon.get(),
        }}


class ExceptionMiddleware():
    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        return self.get_response(request)

    MBD_TEMPLATES = {
        "html-snippet": "mini_buildd/includes/error.html",
        "html": "mini_buildd/error/index.html",
    }

    @classmethod
    def error(cls, request, exception, output="html", status=http.HTTPStatus.INTERNAL_SERVER_ERROR):
        http_exception = util.e2http(exception, status)
        if http_exception.status != http.HTTPStatus.OK:
            # Note: You will not see logs/tracebacks here if not in log-level DEBUG -- otherwise spammers can flood our logs in production
            util.log_exception(LOG, "Middleware exception", exception, level=logging.DEBUG)

        return django.template.response.TemplateResponse(request,
                                                         cls.MBD_TEMPLATES[output],
                                                         {"rfc7807": http_exception.rfc7807},
                                                         status=http_exception.status)

    @classmethod
    def process_exception(cls, request, exception):
        return cls.error(request, exception)

    # Despite having a custom middleware to handle exceptions, the following django default functions still need to be overwritten (see also: urls.py).
    @classmethod
    def bad_request(cls, request, exception):
        return cls.error(request, exception, status=http.HTTPStatus.BAD_REQUEST)

    @classmethod
    def permission_denied(cls, request, exception):
        return cls.error(request, exception, status=http.HTTPStatus.UNAUTHORIZED)

    @classmethod
    def page_not_found(cls, request, exception):
        return cls.error(request, exception, status=http.HTTPStatus.NOT_FOUND)

    @classmethod
    def server_error(cls, request):
        return cls.error(request, None)


class TemplateView(django.views.generic.base.TemplateView):
    """Original django class plus optional custom helpers"""


class DefaultView(TemplateView):
    """Classic view from template; computes template name from request path (``/foo/`` -> ``foo/index.html``, ``/foo/bar.html`` -> ``foo/bar.html``)"""

    def get_template_names(self):
        template = self.request.path.lstrip("/")
        return os.path.join(template, "index.html") if template.endswith("/") else template


class AccountRegisterView(django.views.generic.edit.FormView):
    class TokenGenerator(django.contrib.auth.tokens.PasswordResetTokenGenerator):
        def _make_hash_value(self, user, timestamp):
            return f"{user.pk}{timestamp}{user.is_active}"

    TOKEN_GENERATOR = TokenGenerator()

    class Form(django.contrib.auth.forms.UserCreationForm):
        email = django.forms.EmailField(help_text="Required. User's email address - activation link will be sent there.")

        class Meta:
            model = django.contrib.auth.models.User
            fields = ["username", "email", "password1", "password2"]

    template_name = "accounts/register/index.html"
    form_class = Form

    def form_valid(self, form):
        user = form.save(commit=False)
        user.is_active = False
        user.save()

        url = util.http_endpoint()
        token = self.TOKEN_GENERATOR.make_token(user)
        user.email_user(f"User registration for {url}",
                        (
                            "Dear user,\n\n"
                            f"a registration for user '{user}' has been requested to this email address.\n\n"
                            f"Use this activation link to complete the registration:\n\n"
                            f" {url}/accounts/activate/{django.utils.http.urlsafe_base64_encode(django.utils.encoding.force_bytes(user.pk))}/{token}\n\n"
                            f"Your mini-buildd at {url}\n"
                        ))
        raise util.HTTPOk(f"Activation email sent to: {user.email}")


class AccountActivateView(DefaultView):
    def get(self, request, *args, **kwargs):
        uid = django.utils.encoding.force_str(django.utils.http.urlsafe_base64_decode(kwargs["uidb64"]))
        user = django.contrib.auth.models.User.objects.get(pk=uid)
        if not AccountRegisterView.TOKEN_GENERATOR.check_token(user, kwargs["token"]):
            raise util.HTTPBadRequest("Invalid Activation Token")
        user.is_active = True
        user.save()
        raise util.HTTPOk(f"User '{user}' is now activated")


@method_decorator(login_required, "dispatch")
class AccountProfileView(DefaultView):
    def setup(self, request, *args, **kwargs):
        super().setup(request, *args, **kwargs)
        if self.request.user.is_authenticated:  # fixes error 500 when called logged out: even with login_required, setup() will still be called.
            self.mbd_subscriptions = util.models().Subscription.objects.filter(subscriber=self.request.user)


@method_decorator(login_required, "dispatch")
class AccountNullView(django.views.generic.base.TemplateView):
    """For python client login support only. See :py:meth:`net.ClientEndpoint.login`"""

    def get(self, request, *args, **kwargs):
        return django.http.HttpResponse("<!DOCTYPE html>")


class BuildersView(DefaultView):
    def setup(self, request, *args, **kwargs):
        super().setup(request, *args, **kwargs)
        self.mbd_check = self.request.user.is_authenticated and self.request.user.is_active and self.request.user.is_staff
        self.mbd_builders = util.models().Builders(check=self.mbd_check)


class RouteView(TemplateView):
    def setup(self, request, *args, **kwargs):
        super().setup(request, *args, **kwargs)
        self.mbd_route = config.ROUTES[kwargs["route"]]
        # Be sure to strip leading "/" so consecutive joins won't interpret as new absolute start point (else, any system path may be delivered via an URI like ``foo//any/system/path``)
        self.mbd_path = kwargs.get("path").lstrip("/")
        self.mbd_fullpath = self.mbd_route.path.join(self.mbd_path)
        self.mbd_isfile = os.path.isfile(self.mbd_fullpath)
        self.mbd_static_uri = self.mbd_route.get_uri("static", self.mbd_path)

        if not os.path.exists(self.mbd_fullpath):
            raise util.HTTPNotFound(f"No such path: {self.mbd_path}")


class DirView(RouteView):
    template_name = "mini_buildd/dir.html"

    def setup(self, request, *args, **kwargs):
        super().setup(request, *args, **kwargs)
        self.mbd_dirview = True
        if not self.mbd_isfile:
            self.mbd_scandir = os.scandir(self.mbd_fullpath)


class EmbedView(TemplateView):
    template_name = "mini_buildd/embed.html"


class ManualView(EmbedView):
    def setup(self, request, *args, **kwargs):
        super().setup(request, *args, **kwargs)
        self.mbd_title = "Manual"
        self.mbd_uri = config.URIS["manual"]["static"]


@method_decorator(user_passes_test(lambda u: u.is_superuser), "dispatch")
class SetupView(EmbedView):
    def setup(self, request, *args, **kwargs):
        super().setup(request, *args, **kwargs)
        self.mbd_title = "Setup"
        self.mbd_uri = config.URIS["admin"]["view"]


class EventsView(RouteView):
    template_name = "mini_buildd/events/index.html"

    def setup(self, request, *args, **kwargs):
        super().setup(request, *args, **kwargs)
        # Use events queue if no path is given, else load all events of sub path from files
        if self.mbd_path == "":
            self.mbd_events = daemon.get().events
            self.mbd_title = "Events Queue"
            self.mbd_tooltip = f"Showing {len(self.mbd_events)} (of max. {self.mbd_events.maxlen}) current events (browse events dir or use 'search' for older events)"
        else:
            self.mbd_events = events.load(self.mbd_path)


class RepositoriesView(RouteView):
    @classmethod
    def mbd_parse_distribution(cls, repo, dist, suite):
        name = repo.mbd_get_diststr(dist, suite)
        return {
            "name": name,
            "uri": config.ROUTES["repositories"].uris["static"].join(repo.identity, "dists", name + "/"),
            "mandatory_version_regex": repo.mbd_get_mandatory_version_regex(dist, suite),
            "snapshots": repo.mbd_reprepro.get_snapshots(name),
            "apt_build_sources_list": util.rrpes(lambda dist, suite: repo.mbd_get_apt_build_sources_list(dist, suite).get(with_comment=True), dist, suite),
            "apt_build_preferences": repo.mbd_get_apt_build_preferences(dist, suite),
        }

    def setup(self, request, *args, **kwargs):
        super().setup(request, *args, **kwargs)
        try:
            self.mbd_diststr = pathlib.Path(request.path).parts[5]  # /mini_buildd/repositories/test/dists/<codename>-<repoid>-<suite>[/...]
            self.mbd_repository, self.mbd_distribution, self.mbd_suite = models.parse_diststr(self.mbd_diststr)

            self.mbd_xtra = {}
            for s in self.mbd_repository.layout.suiteoption_set.all():
                self.mbd_xtra[s.suite.name] = self.mbd_parse_distribution(self.mbd_repository, self.mbd_distribution, s)
            self.template_name = "mini_buildd/repositories/distribution.html"
        except Exception as e:
            LOG.debug("No distribution found, showing all repositories: %s", e)
            self.mbd_repositories = util.models().Repository.objects.all
            self.template_name = "mini_buildd/repositories/index.html"


def get_client_ip(request):
    xff = request.META.get("HTTP_X_FORWARDED_FOR")
    return xff.split(",")[0] if xff else request.META.get("REMOTE_ADDR")


class APIView(TemplateView):
    class Snippets(dict):
        """Pre-computed dict of available specialized call templates (just add to or remove files from ``includes/api/``)"""

        def __init__(self):
            super().__init__()
            for name in api.CALLS:
                path = f"mini_buildd/includes/api/{name}.html"
                try:
                    django.template.loader.get_template(path)
                    self[name] = path
                except django.template.TemplateDoesNotExist:
                    self[name] = "mini_buildd/includes/api/default.html"
                LOG.debug("API call '%s' uses HTML snippet: %s", name, self[name])

    MBD_API_SNIPPETS = Snippets()

    def setup(self, request, *args, **kwargs):
        super().setup(request, *args, **kwargs)
        self.mbd_error = None
        self.mbd_output = request.GET.get("output")
        try:
            # Get API call object from class from route
            self.mbd_call = kwargs["call"].from_sloppy_args(**dict(request.GET.items()))
            self.mbd_call.set_request(request)

            # Check authorization
            if not self.mbd_call.AUTH.is_authorized(request.user):
                raise util.HTTPUnauthorized(f"{self.mbd_call} needs authorization: {self.mbd_call.AUTH}")

            # Check if we need a running daemon
            if self.mbd_call.NEEDS_RUNNING_DAEMON and not daemon.get().is_alive():
                raise util.HTTPUnavailable(f"{self.mbd_call} needs running daemon: {self.mbd_call.name()}")

            # Check confirmable calls
            if self.mbd_call.CONFIRM and request.GET.get("confirm", None) != self.mbd_call.name():
                raise util.HTTPBadRequest(f"{self.mbd_call} needs to be confirmed: {self.mbd_call.name()}")

            LOG.info("%s by user '%s' from '%s'", self.mbd_call, request.user, get_client_ip(request))
            self.mbd_call.run()

            self.mbd_api_inc = self.MBD_API_SNIPPETS[self.mbd_call.name()]
            self.template_name = {
                "html-snippet": self.mbd_api_inc,
                "html": "mini_buildd/api/page.html",
            }.get(self.mbd_output)

        except Exception as e:
            util.log_exception(LOG, f"{self.mbd_call} failed", e)
            self.mbd_error = e

    @method_decorator(require_GET)
    def get(self, request, *args, **kwargs):
        # Error handling
        if self.mbd_error is not None:
            if self.mbd_output in ["html", "html-snippet"]:
                return ExceptionMiddleware.error(request, self.mbd_error, output=self.mbd_output)

            http_exception = util.e2http(self.mbd_error)
            return django.http.JsonResponse(http_exception.rfc7807.to_json(),
                                            status=http_exception.status,
                                            content_type="application/problem+json")

        # API call output
        if self.template_name is not None:
            return super().get(request, *args, **kwargs)
        if self.mbd_output == "plain":
            return django.http.HttpResponse(self.mbd_call.plain(), content_type="text/plain")
        return django.http.JsonResponse(self.mbd_call.result, safe=False)



# Generated by Django 4.2.8 on 2024-01-03 16:29

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('mini_buildd', '0006_remove_daemon_custom_hooks_directory'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='aptkey',
            name='auto_reactivate',
        ),
        migrations.RemoveField(
            model_name='chroot',
            name='auto_reactivate',
        ),
        migrations.RemoveField(
            model_name='daemon',
            name='auto_reactivate',
        ),
        migrations.RemoveField(
            model_name='remote',
            name='auto_reactivate',
        ),
        migrations.RemoveField(
            model_name='repository',
            name='auto_reactivate',
        ),
        migrations.RemoveField(
            model_name='source',
            name='auto_reactivate',
        ),
        migrations.RemoveField(
            model_name='uploader',
            name='auto_reactivate',
        ),
    ]



import logging
import threading

import django.conf

from mini_buildd import api, builder, config, cron, events, ftpd, gnupg, packager, threads, util, values

LOG = logging.getLogger(__name__)


class RemotesKeyring(gnupg.TmpGnuPG):
    """Remotes keyring to authorize buildrequests and buildresults"""

    def __init__(self):
        super().__init__(tmpdir_options={"prefix": "gnupg-remotes-keyring-"})
        if get().gnupg.pub_key:
            self.add_pub_key(get().gnupg.pub_key)
        for r in util.models().Remote.mbd_get_active_or_auto_reactivate():
            self.add_pub_key(r.key)
            LOG.debug("Remote key added for '%s': %s: %s", r, r.key_long_id, r.key_name)


class UploadersKeyring(gnupg.TmpGnuPG):
    """Uploader keyring for repository"""

    def __init__(self, repo_identity):
        super().__init__(tmpdir_options={"prefix": f"gnupg-uploaders-keyring-{repo_identity}-"})
        r = util.models().Repository.objects.get(pk=repo_identity)
        # Add keys from django users
        for u in django.contrib.auth.models.User.objects.filter(is_active=True):
            LOG.debug("Checking user: %s", u)
            uploader = None
            try:
                uploader = u.uploader
            except BaseException as e:
                LOG.warning("User '%s' does not have an uploader profile (deliberately removed?): %s", u, e)

            if uploader and uploader.mbd_is_active() and uploader.may_upload_to.all().filter(identity=repo_identity):
                LOG.debug("Adding uploader key for '%s': %s: %s", repo_identity, uploader.key_long_id, uploader.key_name)
                self.add_pub_key(uploader.key)

        # Add configured extra keyrings
        for line in r.extra_uploader_keyrings.splitlines():
            line = line.strip()
            if line and line[0] != "#":
                LOG.debug("Adding keyring: %s", line)
                self.add_keyring(line)

        # Always add our key too for internal builds
        if get().gnupg.pub_key:
            self.add_pub_key(get().gnupg.pub_key)


class Daemon(threads.EventThread):
    def __init__(self):
        self.builder = builder.Builder(max_parallel_builds=get_model().max_parallel_builds)
        self.packager = packager.Packager()
        self.ftpd = ftpd.FtpD(endpoint=get_model().mbd_get_ftp_endpoint())
        self.crontab = cron.Tab()
        super().__init__(subthreads=[self.ftpd, self.packager, self.builder, self.crontab])

        self.events = events.Queue.load(maxlen=get_model().event_queue_size)
        self.gnupg = get_model().mbd_gnupg()
        self.public_key_cache = gnupg.PublicKeyCache()

        self.report = util.attempt(api.Setup.report_load)

    def __str__(self):
        return str(get_model())

    def handshake_message(self):
        return self.gnupg.gpgme_sign(config.HANDSHAKE_MESSAGE)

    def join(self, timeout=None):
        super().join()
        self.events.shutdown()

    def run_event(self, event):
        raise util.HTTPInternal(f"Daemon thread accepts SHUTDOWN event only, got: {event}")

    def get_title(self):
        """Human-readable short title for this Daemon instance"""
        return f"{get_model().identity}@{config.HOSTNAME_FQDN}"

    def mbd_inspect(self):
        util.models().Daemon.Admin.mbd_check(get_model(), force=True)
        inspect = api.Setup.setup_inspect()
        inspect._run()  # pylint: disable=protected-access  # This should not be a "maintenance" call
        self.report = inspect.report
        util.attempt(inspect.report_save)


def get_model():
    model, created = util.models().Daemon.objects.get_or_create(id=1, defaults={"ftpd_bind": config.default_ftp_endpoint()})
    if created:
        LOG.info("New Daemon model instance created")
    return model


_DAEMON = None
_DAEMON_LOCK = threading.Lock()


def init():
    """Call once before you can use get(), start(), stop()"""
    # pylint: disable=global-statement
    global _DAEMON
    _DAEMON = Daemon()


def get():
    with _DAEMON_LOCK:
        return _DAEMON


def start():
    with _DAEMON_LOCK:
        if not get_model().mbd_is_active():
            raise util.HTTPUnavailable("Can't start: Daemon instance is deactivated")
        if not _DAEMON.is_alive():
            values.cache_clear()
            _DAEMON.ftpd.bind(get_model().ftpd_options)
            _DAEMON.start()


def stop():
    with _DAEMON_LOCK:
        if _DAEMON.is_alive():
            _DAEMON.shutdown()
            _DAEMON.join()
        _DAEMON.public_key_cache.close()
        values.cache_clear()
        init()


class Stopped():
    def __init__(self):
        self.start = get().is_alive()

    def __enter__(self):
        stop()

    def __exit__(self, type_, value_, tb_):
        if self.start:
            util.attempt(start)



import ftplib
import glob
import logging
import os
import re
import shutil

import debian.deb822

from mini_buildd import config, daemon, dist, events, files, net, sbuild, util

LOG = logging.getLogger(__name__)


class Base(debian.deb822.Changes):
    def __init__(self, initial_value):
        self.file_path = None
        if isinstance(initial_value, str):
            self.file_path = initial_value
            with util.fopen(initial_value) as cf:
                super().__init__(cf)
        else:
            super().__init__(initial_value)

        self.dfn = files.DebianName(self["Source"], self["Version"])

    def __str__(self):
        """Superclass would show all values multiline, so we need to replace with s.th. less invasive"""
        return self.dfn.changes(self.get("Architecture")) if self.file_path is None else self.file_name()

    def file_name(self):
        return None if self.file_path is None else os.path.basename(self.file_path)

    def dir_path(self):
        return None if self.file_path is None else os.path.dirname(self.file_path)

    def dsc_file_path(self):
        return os.path.join(os.path.dirname(self.file_path), self.dfn.dsc())

    def get_files(self, key=None):
        return [f[key] if key else f for f in self.get("Files", [])]

    def add_file(self, file_name):
        self.setdefault("Files", [])
        self["Files"].append({
            "md5sum": util.Hash(file_name).md5(),
            "size": os.path.getsize(file_name),
            "section": "mini-buildd",
            "priority": "extra",
            "name": os.path.basename(file_name),
        })

    def upload_file_path(self):
        return os.path.splitext(self.file_path)[0] + ".upload"

    def uploaded(self):
        return os.path.exists(self.upload_file_path())

    def _upload(self, ftp_cls, endpoint, force=False):
        pre_upload_timestamp = util.Datetime.now()

        upload = self.upload_file_path()
        if not force and os.path.exists(upload):
            with util.fopen(upload) as uf:
                LOG.warning("FTP: '%s' exists (already uploaded to '%s')", upload, uf.read())
        else:
            with ftp_cls() as ftp:
                ftp.connect(endpoint.options.get("host"), int(endpoint.options.get("port")))
                if isinstance(ftp, ftplib.FTP_TLS):
                    ftp.auth()
                    ftp.prot_p()
                ftp.login()
                ftp.cwd("/incoming")
                for fd in self.get_files() + [{"name": self.file_name()}]:
                    f = fd["name"]
                    LOG.debug("FTP: Uploading file: '%s'", f)
                    with open(os.path.join(os.path.dirname(self.file_path), f), "rb") as fi:
                        ftp.storbinary(f"STOR {f}", fi)
                with util.fopen(upload, "w") as fi:
                    fi.write(endpoint.geturl())
                LOG.info("FTP: '%s' uploaded to '%s'.", self.file_name(), endpoint.geturl())

        return pre_upload_timestamp  # for events --after=

    def upload(self, endpoint, force=False):
        return self._upload(ftplib.FTP_TLS if endpoint.is_ssl() else ftplib.FTP, endpoint, force=force)

    def tar(self):
        _files = [os.path.join(os.path.dirname(self.file_path), f) for f in self.get_files("name")] + [self.file_path]
        return files.Tar().add(_files)

    def untar(self, dir_path):
        files.Tar.extract(self.file_path + ".tar", dir_path)

    def save_as(self, file_path, gnupg=None):
        """Write to file (optionally signed)"""
        LOG.debug("Saving changes: %s", file_path)
        os.makedirs(os.path.dirname(file_path), exist_ok=True)  # Be sure base dir is always available
        with open(file_path, "w+", encoding=config.CHAR_ENCODING) as f:
            f.write(self.dump())
        if gnupg is not None:
            LOG.debug("Signing changes: %s", file_path)
            gnupg.sign(file_path)
        self.file_path = file_path

    def _move_to(self, dir_path, mv):
        """Move (or copy) changes file along with all accompanying files (needs ``file_path`` to be non-None)"""
        LOG.debug("%s changes: '%s'->'%s'", mv.__name__, self.file_path, dir_path)
        os.makedirs(dir_path, exist_ok=True)
        for fd in [{"name": self.file_name()}] + self.get_files():
            f = fd["name"]
            f_abs = os.path.join(self.dir_path(), f)
            LOG.debug("%s '%s' to '%s'", mv.__name__, f, dir_path)
            mv(f_abs, os.path.join(dir_path, f))
        self.file_path = os.path.join(dir_path, self.file_name())

    def move_to(self, path):
        return self._move_to(path, mv=os.rename)

    def copy_to(self, path):
        return self._move_to(path, mv=shutil.copy)


class Changes(Base):
    FILE_APX = ""
    #: <key>: <public_hint>
    MANDATORY = {"Distribution": "", "Source": "", "Version": "", "Architecture": ""}

    @classmethod
    def match(cls, file_path):
        return re.compile(f"^.+{files.DebianName.uext(cls.FILE_APX)}_[^_]+.changes$").match(file_path)

    def __init__(self, initial_value):
        super().__init__(initial_value)

        for key, hint in self.MANDATORY.items():
            if key not in self:
                hint_apx = f" ({hint})" if hint else ""
                raise util.HTTPBadRequest(f"Changes missing mandatory key: {key}{hint_apx}")

        self._timecode = util.Datetime.timecode()

        #: Key to identify if a 'Package' (Upload, Buildrequest, Buildresult).
        self.key = os.path.join(self["Source"], self["Version"], self.timecode)
        #: Key to identify a 'Build' (Buildrequest, Buildresult).
        self.bkey = os.path.join(self.key, self["Architecture"])

    def get_events_path(self, *args):
        return os.path.join(config.ROUTES["events"].path.full, self.bkey, *args)

    @property
    def timecode(self):
        return self.cget("Timecode", self._timecode)

    def cget(self, field, default=None):
        return self.get(util.CField(field).fullname, default=default)

    def cset(self, field, value):
        self[util.CField(field).fullname] = value

    def fields(self):
        return {util.Field(field): value for field, value in self.items()}

    def to_event_json(self):
        return {
            **{
                "timecode": self.timecode,
                "key": self.key,
                "bkey": self.bkey,
                "changes": os.path.join(self.bkey, self.file_name()) if self.file_name() else None,
            },
            **{field.snake_name: value for field, value in self.fields().items() if field.is_cfield or field.snake_name in ["architecture", "maintainer", "changed_by"]},
        }

    def save_to(self, dir_path, tar=None):
        os.makedirs(dir_path, exist_ok=True)
        file_path = os.path.join(dir_path, self.dfn.changes(self["Architecture"], self.FILE_APX))
        if tar is not None:
            tar_path = file_path + ".tar"
            tar.save_as(tar_path)
            self.add_file(tar_path)
        self.save_as(file_path, daemon.get().gnupg)


class Upload(Changes):
    class Options():
        """
        Uploader options in changes

        >>> config.ROUTES = config.Routes("./test-data/home")   # We need ROUTES setup

        >>> f"{Upload.Options(Base('test-data/changes.options'))}"
        "auto-ports=['jessie-test-unstable', 'squeeze-test-snasphot'], autopkgtest-mode=errfail, internal-apt-priority=543, lintian-mode=errfail, piuparts-mode=errfail"

        >>> f"{Upload.Options(Base('test-data/changes.options_1.0'))}"
        "auto-ports=['jessie-test-unstable', 'squeeze-test-snasphot'], internal-apt-priority=543, lintian-mode=errfail, lintian-mode[i386]=disabled"

        >>> f"{Upload.Options(Base('test-data/changes.options_0.8'))}"
        "auto-ports=['jessie-test-unstable', 'squeeze-test-snasphot'], lintian-mode=ignore"
        """

        class Bool():
            _TRUE = ["true", "1"]
            _FALSE = ["false", "0"]
            _VALID = _TRUE + _FALSE

            def __init__(self, raw_value):
                if raw_value.lower() not in self._VALID:
                    raise util.HTTPBadRequest(f"Bool value must be one of {','.join(self._VALID)}")
                self.value = raw_value.lower() in self._TRUE

        class Int():
            def __init__(self, raw_value):
                self.value = int(raw_value)

        class Str():
            def __init__(self, raw_value):
                self.value = str(raw_value)

        class CSV():
            def __init__(self, raw_value):
                self.value = raw_value.split(",")

        KEYWORD = "MINI_BUILDD_OPTION"

        _OPTIONS_DEPRECATED = {
            "ignore-lintian": Bool,
            "run-lintian": Bool,
        }

        _OPTIONS = {**{
            "lintian-mode": Str,
            "piuparts-mode": Str,
            "autopkgtest-mode": Str,
            "internal-apt-priority": Int,
            "auto-ports": CSV,
            "deb-build-options": Str,
            "deb-build-profiles": Str,
            "add-depends": Str,
        }, **_OPTIONS_DEPRECATED}

        @classmethod
        def usage(cls):
            return (
                "(lintian|piuparts|autopkgtest)-mode[<arch>]=<mode>\n"
                f"  {dist.SbuildCheck.usage()}\n"
                "internal-apt-priority[<arch>]=<prio>\n"
                "  Apt prio for our repo packages when building (use 500 to always prefer).\n"
                "auto-ports=<dist>,...\n"
                "  List of distributions (comma-separated) to automatically run ports for after successful install.\n"
                "deb-build-options[<arch>]=<option> <option>...\n"
                "  Contents of DEB_BUILD_OPTIONS environment for building (see https://www.debian.org/doc/debian-policy/ch-source.html#debian-rules-and-deb-build-options).\n"
                "deb-build-profiles[<arch>]=<profile> <profile>...\n"
                "  Contents of DEB_BUILD_PROFILES environment for building (see https://wiki.debian.org/BuildProfileSpec).\n"
                "add-depends[<arch>]=<dep>,<dep>...\n"
                "  List of additional dependencies (comma-separated, like in ``debian/control``) (see ``sbuild --add-depends``).\n"
            )

        @classmethod
        def _get_top_changes(cls, upload_changes):
            """
            Filter only the first block from the changes (changelog) entry

            Upload changes may include multiple version blocks from
            the changelog (internal porting does it, for example),
            but we must only consider values from the top one.
            """
            result = ""
            header_found = False
            for line in upload_changes.get("Changes", "").splitlines(True):
                if re.match(r"^ [a-z0-9]+", line):
                    if header_found:
                        break
                    header_found = True
                result += line
            return result

        def _parse_ancient(self):
            """Compat parse support for ancient 'magic' options"""
            def warning(ancient, option):
                LOG.warning("Ancient user option \"%s\" found. Please use option \"%s\" instead (see user manual).", ancient, option)

            ancient_auto_backports = re.search(r"\*\s*MINI_BUILDD:\s*AUTO_BACKPORTS:\s*([^*.\[\]]+)", self._top_changes)
            if ancient_auto_backports:
                warning("AUTO_BACKPORTS", "auto-ports")
                self._set("auto-ports", ancient_auto_backports.group(1))

            ancient_backport_mode = re.search(r"\*\s*MINI_BUILDD:\s*BACKPORT_MODE", self._top_changes)
            if ancient_backport_mode:
                warning("BACKPORT_MODE", "lintian-mode")
                self._set("lintian-mode", "ignore")

        def __init__(self, upload_changes):
            self._top_changes = self._get_top_changes(upload_changes)
            self._options = {}
            matches = re.findall(fr"\*\s*{self.KEYWORD}:\s*([^*.]+)=([^*.]+)", self._top_changes)
            for m in matches:
                self._set(m[0], m[1])

            self._parse_ancient()

        def __str__(self):
            return ", ".join(f"{key}={value}" for key, value in sorted(self._options.items()))

        def _set(self, key, raw_value):
            key_part = key.partition("[")
            base_key = key_part[0]
            value = re.sub(r"\s+", "", raw_value)

            # Validity check for key
            if base_key not in list(self._OPTIONS):
                raise util.HTTPBadRequest(f"Unknown upload option: {key}")

            # Duplicity check
            if key in list(self._options):
                raise util.HTTPBadRequest(f"Duplicate upload option: {key}")

            # Value conversion check
            converted_value = None
            try:
                converted_value = self._OPTIONS[base_key](value)
            except BaseException as e:
                raise util.HTTPBadRequest(f"Invalid upload option value: {key}=\"{value}\"") from e

            # Handle deprecated options
            if base_key in list(self._OPTIONS_DEPRECATED):
                oldkey = key
                oldvalue = converted_value.value
                alt = key_part[1] + key_part[2]
                if base_key == "ignore-lintian":
                    key = f"lintian-mode{alt}"
                    converted_value.value = "ignore" if oldvalue else "errfail"
                elif base_key == "run-lintian":
                    key = f"lintian-mode{alt}"
                    converted_value.value = "errfail" if oldvalue else "disabled"
                LOG.warning("User Option: Converted DEPRECATED %s=%s to %s=%s", oldkey, oldvalue, key, converted_value.value)

            self._options[key] = converted_value.value

            LOG.debug("Upload option set: %s=\"%s\"", key, value)

        def get(self, key, alt=None, default=None):
            """Get first existing option value in this order: key[a], key, default"""
            # Validity check for key
            if key not in list(self._OPTIONS):
                raise util.HTTPInternal(f"Unknown key used for get(): {key}")

            if alt:
                m_key = f"{key}[{alt}]"
                if m_key in self._options:
                    return self._options.get(m_key, default)
            return self._options.get(key, default)

    MANDATORY = {**Changes.MANDATORY, **{"Changed-By": "Maybe the changelog entry was not finalised?"}}

    def __init__(self, initial_value, **kwargs):
        super().__init__(initial_value, **kwargs)
        self.options = self.Options(self)
        # (Optionally map and) check distribution string integrity
        self.dist = dist.Dist(util.models().map_distribution(self["Distribution"]))

    def check_mode(self, checker, alt, fallback_mode_int):
        """Get check mode from upload options, with fallback"""
        options_mode_str = self.options.get(f"{checker}-mode", alt=alt)
        fallback_mode_str = dist.SbuildCheck.Mode(fallback_mode_int).name
        return dist.SbuildCheck(checker, options_mode_str if options_mode_str is not None else fallback_mode_str)

    def request_builds(self, repository, distribution, suite_option):
        """Build buildrequest files for all architectures"""
        try:
            with util.fopen(self.dsc_file_path()) as dsc_file:
                dsc = debian.deb822.Dsc(dsc_file)
        except FileNotFoundError as e:
            raise util.HTTPBadRequest(f"Missing DSC file '{self.dfn.dsc()}' (Maybe ``dpkg-buildpackage`` was called with '-b/-B'?)") from e

        # Get all DSC/source package files:
        # - Check md5 against possible pool files.
        # - Add missing from pool (i.e., orig.tar.gz).
        # - make sure all files from dsc are actually available
        files_from_dsc = []
        for f in dsc["Files"]:
            from_pool = None
            for p in glob.glob(os.path.join(repository.mbd_path, "pool", "*", "*", self["Source"], f["name"])):
                if f["md5sum"] != util.Hash(p).md5():
                    raise util.HTTPBadRequest(f"MD5 mismatch in uploaded DSC vs. pool: {f['name']}")
                from_pool = p

            if f["name"] in self.get_files(key="name"):
                files_from_dsc.append(os.path.join(os.path.dirname(self.dsc_file_path()), f["name"]))
            elif from_pool is not None:
                files_from_dsc.append(from_pool)
                LOG.info("%s: File added from pool: %s", self, from_pool)
            else:
                raise util.HTTPBadRequest(f"Missing file '{f['name']}' neither in upload, nor in pool (use '-sa' for uploads with new upstream)")

        buildrequest_dict = {}
        for ao in distribution.architectureoption_set.all():
            path = os.path.join(self.dir_path(), ao.architecture.name)   # Usually: <source>/<version>/<timecode>/source/<arch>/
            os.makedirs(path, exist_ok=True)

            buildrequest = Buildrequest({
                "Distribution": self.dist.get(),
                "Source": self["Source"],
                "Version": self["Version"],
                "Architecture": ao.architecture.name,
                util.CField("Timecode").fullname: self.timecode,
            })

            # Generate config files
            config_path = os.path.join(path, sbuild.CONFIG_DIR)
            os.makedirs(config_path, exist_ok=True)
            config_files = {
                sbuild.CONFIG_APT_SOURCES_LIST: repository.mbd_get_apt_build_sources_list(distribution, suite_option).get(),
                sbuild.CONFIG_APT_PREFERENCES: repository.mbd_get_apt_build_preferences(distribution, suite_option, self.options.get("internal-apt-priority")),
                sbuild.CONFIG_APT_KEYS: repository.mbd_get_apt_keys(distribution),
                sbuild.CONFIG_SSL_CERT: util.http_endpoint().get_certificate(),
                sbuild.CONFIG_CHROOT_SETUP_SCRIPT: distribution.chroot_setup_script,
                sbuild.CONFIG_SBUILDRC_SNIPPET: distribution.sbuildrc_snippet,
            }
            for file_name, content in config_files.items():
                with util.fopen(os.path.join(config_path, file_name), "w") as f:
                    f.write(content)

            # Generate tar
            tar = files.Tar()
            tar.add([os.path.join(config_path, file_name) for file_name in config_files], arcdir=sbuild.CONFIG_DIR)
            tar.add([self.dsc_file_path()])
            tar.add(files_from_dsc)

            buildrequest.cset("Upload-To", daemon.get_model().mbd_get_ftp_endpoint().geturl())
            buildrequest.cset("Packager", util.http_endpoint().geturl())
            if ao.build_architecture_all:
                buildrequest.cset("Arch-All", "Yes")
            buildrequest.cset("Build-Dep-Resolver", distribution.get_build_dep_resolver_display())

            buildrequest.cset(sbuild.CONFIG_BLOCKS.name, distribution.sbuild_config_blocks)
            buildrequest.cset(sbuild.SETUP_BLOCKS.name, distribution.sbuild_setup_blocks)

            buildrequest.cset("Deb-Build-Options", self.options.get("deb-build-options", ao.architecture.name, distribution.deb_build_options))
            buildrequest.cset("Deb-Build-Profiles", self.options.get("deb-build-profiles", ao.architecture.name, distribution.deb_build_profiles))
            buildrequest.cset("Add-Depends", self.options.get("add-depends", ao.architecture.name, distribution.add_depends))

            buildrequest.cset("Lintian-Mode", self.check_mode("lintian", ao.architecture.name, distribution.lintian_mode).mode.name)
            buildrequest.cset("Lintian-Extra-Options", distribution.lintian_extra_options)
            buildrequest.cset("Lintian-Warnfail-Options", distribution.lintian_warnfail_options)
            buildrequest.cset("Autopkgtest-Mode", self.check_mode("autopkgtest", ao.architecture.name, distribution.autopkgtest_mode).mode.name)
            buildrequest.cset("Piuparts-Mode", self.check_mode("piuparts", ao.architecture.name, distribution.piuparts_mode).mode.name)

            buildrequest.cset("Buildlog", os.path.join(buildrequest.bkey, self.dfn.buildlog(ao.architecture.name)))

            builders = util.models().Builders(wake=True, check=True).get(self.dist.codename, ao.architecture.name)
            for _load, builder in sorted(builders.items()):
                try:
                    buildrequest.cset("Builder", builder["url"])
                    buildrequest.cset("Buildlog-Building", config.URIS["builds"]["dir"].url_join(buildrequest.cget("Buildlog"), endpoint=net.ClientEndpoint(builder["url"])))
                    buildrequest.cset("Buildlog-Installed", config.URIS["events"]["dir"].url_join(buildrequest.cget("Buildlog")))
                    buildrequest.save_to(self.get_events_path(ao.architecture.name), tar)
                    buildrequest.upload(net.ClientEndpoint(builder["incoming_url"]))
                    buildrequest_dict[ao.architecture.name] = buildrequest
                    break
                except BaseException as e:
                    util.log_exception(LOG, f"Uploading to '{builder['incoming_url']}' failed", e, level=logging.ERROR)
            if ao.architecture.name not in buildrequest_dict:
                raise util.HTTPUnavailable(f"No working builder found for {ao.architecture.name}/{self.dist.codename}")
        return buildrequest_dict


class Buildrequest(Changes):
    FILE_APX = "mini-buildd-buildrequest"

    def __init__(self, initial_value, **kwargs):
        super().__init__(initial_value, **kwargs)
        self.dist = dist.Dist(self["Distribution"])

    def check_mode(self, checker):
        """Get check mode from checker (lintian, piuparts, autopkgtest)"""
        return dist.SbuildCheck(checker, self.cget(f"{checker.capitalize()}-Mode", dist.SbuildCheck.Mode.DISABLED.name))

    def check_extra_options(self, checker):
        """Get check mode from checker (lintian, piuparts, autopkgtest)"""
        return self.cget(f"{checker.capitalize()}-Extra-Options", "")

    def gen_buildresult(self):
        bres_path = self.dfn.changes(self["Architecture"], Buildresult.FILE_APX)
        if os.path.exists(bres_path):
            return Buildresult(bres_path)
        return Buildresult({k: self[k] for k in ["Distribution", "Source", "Version", "Architecture"] + [util.CField(cfield).fullname for cfield in ["Timecode", "Builder", "Buildlog", "Buildlog-Installed", "Upload-To"]]})

    def schroot_name(self):
        """
        Schroot name w/o namespace (see ``schroot --list``)

        Must produce same as :func:`~models.chroot.Chroot.mbd_schroot_name`.
        """
        return f"mini-buildd-{self.dist.codename}-{self['Architecture']}"


class Buildresult(Changes):
    FILE_APX = "mini-buildd-buildresult"

    def success(self, upload, distribution, ignore_checks):
        arch = self["Architecture"]
        status = self.cget("Sbuild-Status")
        lintian = self.cget("Sbuild-Lintian")
        piuparts = self.cget("Sbuild-Piuparts")
        autopkgtest = self.cget("Sbuild-Autopkgtest")

        def check(arch, status, checker, mode):
            """
            Check if arch is allowed to succeed based on check results

            Succeeds if
            - the actual check run has no errors or warnings
            - the distribution is configured to ignore check results
            - upload says to ignore check results
            - suite is experimental (we deliberately ignore check in that case)
            - warnings are present, but check mode is not warnfail
            """
            check_mode = upload.check_mode(checker, arch, mode)
            return check_mode.check(status, ignore=ignore_checks)

        def check_all():
            return \
                check(arch, lintian, "lintian", distribution.lintian_mode) and \
                check(arch, piuparts, "piuparts", distribution.piuparts_mode) and \
                check(arch, autopkgtest, "autopkgtest", distribution.autopkgtest_mode)

        return check_all() if status == "successful" else status == "skipped"


def incoming2queue(file_path):
    LOG.debug("Trying to queue incoming: %s", file_path)
    try:
        if Buildresult.match(file_path):
            changes = Buildresult(file_path)
            changes.move_to(changes.get_events_path())
            daemon.get().packager.queue.put(changes)
        elif Buildrequest.match(file_path):
            changes = Buildrequest(file_path)
            changes.move_to(config.ROUTES["builds"].path.join(changes.bkey))
            daemon.get().builder.queue.put(changes)
        else:  # User Upload
            changes = Upload(file_path)
            changes.move_to(changes.get_events_path())
            daemon.get().packager.queue.put(changes)
    except Exception as e:
        # If it at least qualifies as Changes, still try error/event handling for user experience
        changes = Changes(file_path)
        changes.move_to(changes.get_events_path())
        daemon.get().events.log(events.Type.REJECTED, changes, e)



# Generated by Django 4.2.9 on 2024-01-11 16:43

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('mini_buildd', '0024_remove_daemon_keep_build_results_and_more'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='daemon',
            name='show_last_builds',
        ),
    ]



import collections
import contextlib
import glob
import logging
import os
import shlex
import shutil
import stat
import subprocess
import tarfile

from mini_buildd import call, util

LOG = logging.getLogger(__name__)


def find(path, name):
    """
    Find file name in path, recursively -- return absolute path of first hit, else None
    """
    def _glob(path, name):
        result = glob.glob(os.path.join(path, "**", name), recursive=True)
        return result[0] if result else None

    def _scandir(path, name):
        for entry in os.scandir(path):
            if entry.is_file() and entry.name == name:
                yield entry.path
            elif entry.is_dir(follow_symlinks=False):
                yield from _scandir(entry.path, name)

    def _walk(path, name):
        for root, _dirs, files in os.walk(path):
            if name in files:
                return os.path.join(root, name)
        return None

    return _walk(path, name)


class DebianName:
    """
    Debian package-related file names, based on package ``name`` (usually source name) and ``version``

    Always strip epoch from version, and handles special
    mini-buildd types.

    >>> dfn = DebianName("mypkg", "7:1.2.3-1")
    >>> dfn.dsc()
    'mypkg_1.2.3-1.dsc'
    >>> dfn.changes("amd64")
    'mypkg_1.2.3-1_amd64.changes'
    >>> dfn.changes("mips")
    'mypkg_1.2.3-1_mips.changes'
    >>> dfn.changes("mips", apx="my-custom-apx")
    'mypkg_1.2.3-1_my-custom-apx_mips.changes'
    >>> dfn.buildlog("mips")
    'mypkg_1.2.3-1_mips.buildlog'
    """

    def __init__(self, package, version):
        self.base = f"{package}_{util.strip_epoch(version)}"

    @classmethod
    def uext(cls, ext):
        return f"_{ext}" if ext else ""

    def gen(self, type_, apx=None, arch=None):
        return f"{self.base}{self.uext(apx)}{self.uext(arch)}.{type_}"

    def dsc(self):
        return self.gen("dsc")

    def changes(self, arch, apx=None):
        return self.gen("changes", apx=apx, arch=arch)

    def upload(self, arch, apx=None):
        return self.gen("upload", apx=apx, arch=arch)

    def buildlog(self, arch):
        return self.gen("buildlog", arch=arch)

    def json(self, apx, arch=None):
        return self.gen("json", apx=apx, arch=arch)


class Tar():
    def __init__(self):
        self.files = []

    def add(self, file_paths, arcdir=""):
        self.files.append((arcdir, file_paths))
        return self

    def save_as(self, file_path):
        with contextlib.closing(tarfile.open(file_path, "w")) as t:
            for arcdir, files in self.files:
                for f in files:
                    t.add(f, arcname=os.path.join(arcdir, os.path.basename(f)))

    @classmethod
    def extract(cls, file_path, dest_path):
        with contextlib.closing(tarfile.open(file_path, "r")) as t:
            t.extractall(path=dest_path)


class Path:
    """
    PATH string with some convenience functionality

    >>> p = Path("/tmp", ["uff", None, "tata"])
    >>> p.base
    '/tmp'
    >>> p.sub
    'uff/tata'
    >>> p.full
    '/tmp/uff/tata'
    >>> p.join("more", "sub", "dirs")
    '/tmp/uff/tata/more/sub/dirs'

    >>> p = Path("/tmp")
    >>> p.sub
    ''
    """

    def __init__(self, base, subdirs=None, create=False):
        if len(base) < 4:
            raise util.HTTPInternal(f"Possibly dangerous path '{base}'")  # Be a bit paranoid since path.join() results are used to expire-purge things
        self.base = base
        self.sub = os.path.join("", *[s for s in ([] if subdirs is None else subdirs) if s is not None])
        self.full = os.path.join(self.base, self.sub)
        if create:
            self.create()
        self.exists = os.path.isdir(self.full)

    def create(self):
        os.makedirs(self.full, exist_ok=True)

    def new_sub(self, subdirs=None, create=False):
        return self.__class__(self.full, subdirs, create=create)

    def __str__(self):
        return self.full

    def join(self, *args):
        return os.path.join(self.full, *args)

    def sub_join(self, *args):
        return os.path.join(self.sub, *args)

    def expire(self, older_than):
        """Expire directories 'older than'"""
        valid_until = util.Datetime.now() - older_than
        LOG.debug("FS expire items below '%s' older than %s...", self.full, older_than)
        for item in glob.glob(self.join("**/*"), recursive=True):
            if os.path.isdir(item):
                LOG.debug("FS expire check older than %s on dir: %s", older_than, item)
                if util.Datetime.from_path(item) < valid_until:
                    shutil.rmtree(item)
                    LOG.info("Directory older than %s expired: %s", older_than, item)

    def ifiles(self, path, extension, after=None, before=None):
        """Fast recursive file finder using os.scandir()"""
        for entry in os.scandir(self.join(path)):
            if entry.is_file() and entry.name.endswith(extension):
                stamp = util.Datetime.from_stamp(entry.stat().st_mtime)
                if (after is None or stamp > after) and (before is None or stamp < before):
                    yield entry
            elif entry.is_dir(follow_symlinks=False):
                yield from self.ifiles(entry.path, extension, after, before)

    def removeprefix(self, path):
        return path.removeprefix(self.full)


class File(collections.deque):
    r"""
    Line based text file (config or script)

    >>> pm = PerlModule("perl").add("# set foo\n$bar = undef;")
    >>> pm.get()
    '# set foo\n$bar = undef;\n1;\n'
    >>> pm.get(snippet=True)
    '# set foo\n$bar = undef;\n'
    >>> pm.get(snippet=True, comments=False)
    '$bar = undef;\n'

    >>> bs = BashScript("bash").add("# echo foo\necho bar")
    >>> bs.get()
    '#!/bin/bash -e\n# echo foo\necho bar\n'
    >>> bs.get(snippet=True)
    '# echo foo\necho bar\n'
    >>> bs.get(snippet=True, comments=False)
    'echo bar\n'

    >>> test_file = "/tmp/mini_buildd_test_conf_file"
    >>> if os.path.exists(test_file): os.remove(test_file)
    >>> File("test_file").add("my_option=7\n").update_as(test_file)
    True
    >>> File("test_file").add("my_option=7\n").update_as(test_file)
    False
    >>> File("test_file").add("my_option=8\n").update_as(test_file)
    True

    >>> dos = File("dos.txt").add_file("test-data/dos.txt")
    >>> with util.fopen("test-data/unix.txt", newline="") as unix: unix.read() == dos.get()
    True
    >>> with util.fopen("test-data/dos.txt", newline="") as dos_file: dos_file.read() == dos.get()
    False

    # rmtree: Enable with caution
    # >>> dir = Dir("./test-data/dirtest").add(File("file1.txt", snippet="test1 data")).add(File("file2.txt", snippet="test2 data"))
    # >>> import shutil
    # >>> shutil.rmtree(dir.path, ignore_errors=True)
    # >>> dir.update()
    # True
    # >>> dir.update()
    # False
    # >>> dir.update()
    # False
    # >>> dummy = dir["file1.txt"].add("nonne zeile")
    # >>> dir.update()
    # True
    # >>> shutil.rmtree(dir.path)
    """

    @classmethod
    def shebang_from_data(cls, data):
        first_line = data.partition("\n")[0]
        if first_line[:2] == "#!":
            return first_line.rstrip()
        raise util.HTTPBadRequest(f"Not a shebang: '{first_line}'")

    @classmethod
    def shebang_from_path(cls, path):
        with util.fopen(path) as f:
            return cls.shebang_from_data(f.readline())

    def __init__(self, name, description=None, snippet=None, comment="#", shebang=None, finalizer=None, placeholders=None):
        super().__init__()
        self.name, self.description, self.comment, self.shebang, self.finalizer, self.placeholders = name, description, comment, shebang, finalizer, placeholders
        if description:
            self.append(self.banner(description))
        if snippet:
            self.add(snippet)

    def __str__(self):
        return f"{self.name}: {self.description}"

    def get(self, snippet=False, comments=True):
        result = ""
        if not snippet and self.shebang:
            result += self.shebang + "\n"
        for line in self:
            if line and line[0] == self.comment and not comments:
                continue
            result += line + "\n"
        if not snippet and self.finalizer:
            result += self.finalizer + "\n"

        return util.subst_placeholders(result, self.placeholders) if self.placeholders is not None else result

    def banner(self, description):
        return f"{self.comment} {self.name}: {description}"

    def add(self, data, snippet=True, description=None):
        if description:
            self.append(self.banner(description))

        lines = data.splitlines()
        if lines:
            if not snippet and self.shebang is not None:
                if lines[0] != self.shebang:
                    raise util.HTTPBadRequest(f"Wrong shebang '{lines[0]}' (needs '{self.shebang}')")
                del lines[0]
            if not snippet and self.finalizer is not None:
                if lines[-1] != self.finalizer:
                    raise util.HTTPBadRequest(f"Wrong finalizer '{lines[-1]}' (needs '{self.finalizer}')")
                del lines[-1]
            for line in lines:
                self.append(line)
        return self

    def add_file(self, path, **kwargs):
        with util.fopen(path) as f:
            self.add(f.read(), **kwargs)
        return self

    def save_as(self, path, **kwargs):
        with util.fopen(path, "w+") as f:
            content = self.get(**kwargs)
            f.write(content)
            if self.shebang is not None:
                os.chmod(path, os.stat(path).st_mode | stat.S_IEXEC)
            return content

    def update_as(self, path, **kwargs):
        """Like save_as(), but return bool whether file has changed on disk"""
        content_pre = None
        if os.path.exists(path):
            with util.fopen(path, "r") as f:
                content_pre = f.read()
        content = self.save_as(path, **kwargs)
        return content_pre != content

    def save(self, dir_, **kwargs):
        return self.save_as(os.path.join(dir_, self.name), **kwargs)

    def update(self, dir_, **kwargs):
        return self.update_as(os.path.join(dir_, self.name), **kwargs)

    def run(self, cli=False):
        with util.tmp_dir(prefix="script-") as tmpdir:
            script_file = os.path.join(tmpdir, self.name)
            LOG.debug("Executing script '%s' from '%s'", self.name, script_file)
            self.save_as(script_file)
            if cli:
                subprocess.check_call([script_file])
            else:
                call.Call([script_file]).check()


class Script(File):
    """Load a complete script file as ``File`` (with automatic shebang recognition)"""

    def __init__(self, script_path):
        with util.fopen(script_path) as f:
            content = f.read()
            super().__init__(os.path.basename(script_path), shebang=self.shebang_from_data(content))
            self.add(content, snippet=False)


class PerlModule(File):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, finalizer="1;", **kwargs)


class ShellScript(File):
    def banner(self, description):
        result = f"printf 'MINI_BUILDD: %s: %s\\n' {shlex.quote(self.name)} {shlex.quote(description)};"
        return result


class ShScript(ShellScript):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, shebang="#!/bin/sh -e", **kwargs)


class BashScript(ShellScript):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, shebang="#!/bin/bash -e", **kwargs)


class AutoScript(File):
    def __init__(self, path, **kwargs):
        super().__init__(os.path.basename(path), shebang=self.shebang_from_path(path), **kwargs)
        self.add_file(path, snippet=False)


class Dir(collections.OrderedDict):
    def __init__(self, path=None, enumerate_file_names=-1):
        super().__init__()
        self.path = path
        self.next_no = enumerate_file_names

    def add(self, file_):
        file_name = f"{self.next_no:03}_{file_.name}" if self.next_no > -1 else file_.name

        if file_name in self.keys():
            raise util.HTTPBadRequest(f"File name {file_name} already exists in Dir object")

        self[file_name] = file_

        if self.next_no > -1:
            self.next_no += 1

        return self

    def _path(self, path=None):
        path = self.path if path is None else path
        if path is None:
            raise util.HTTPBadRequest(f"No directory given to save '{self}' to")
        os.makedirs(path, exist_ok=True)
        return path

    def save(self, path=None):
        path = self._path(path)
        for file_name, file_ in self.items():
            file_.save_as(os.path.join(path, file_name))

    def update(self, path=None):
        changed = set()
        path = self._path(path)
        for file_ in self.values():
            changed.add(file_.update(path))
        return True in changed


class AptLine:
    def __init__(self, mirror, path, suite, components, options=None, comment=None):
        """
        APT line

        See ``man 5 sources.list`` (arg nomenclature taken from there; 'uri' is 'mirror + path').

        >>> l0 = AptLine("http://localhost", "debian", "dist", ["c1", "c3", "c2"])
        >>> l0.get()
        'deb [] http://localhost/debian dist c1 c3 c2'
        >>> l1 = AptLine("http://localhost", "debian", "dist", ["c1", "c2"], options={"trusted": "true", "check-valid-until": "false"})
        >>> l1.get()
        'deb [trusted=true check-valid-until=false] http://localhost/debian dist c1 c2'
        >>> l2 = AptLine("http://localhost", "debian", "dist", ["c1", "c2"])
        >>> l2.get("deb-src")
        'deb-src [] http://localhost/debian dist c1 c2'
        >>> l3 = AptLine("http://localhost", "debian", "dist", [])  # Test trailing "/" if no components
        >>> l3.get("deb")
        'deb [] http://localhost/debian dist/'
        >>> SourcesList([l0, l1, l2]).get(types=["deb", "deb-src"])  # doctest: +ELLIPSIS
        'deb ...
        """
        self.mirror = mirror
        self.path = path
        self.suite = suite
        self.components = " ".join(components)
        self.options = options if isinstance(options, str) else " ".join((f'{o}={v}' for o, v in ({} if options is None else options).items()))
        self.comment = comment

    def get(self, type_="deb", mirror=None, options=None):
        suite_quote = '"' if ' ' in self.suite else ''
        return (
            f"{type_} "
            f"[{self.options}{' ' if options and self.options else ''}{'' if options is None else options}] "
            f"{os.path.join(mirror or self.mirror, self.path)} "
            f"{suite_quote}{self.suite}{f'{suite_quote} ' if self.components else f'/{suite_quote}'}"
            f"{self.components}"
        )


class SourcesList(list):
    def get(self, types=None, mirror=None, options=None, with_comment=False):
        return "\n".join((f"{f'# {line.comment}{util.NEWLINE}' if with_comment else ''}{line.get(type_, mirror=mirror, options=options)}" for line in self for type_ in (["deb"] if types is None else types))) + "\n"
